--[[
		##########################################
		##	5iver Survival Land					##
		##										##
		##	Thanks to all survival				##
		##	script makers, contributors, GPG 	##
		##	forums and especially GPG devs	 	##
		##########################################
]]--

--LOG("_____LOG: Starting Script")
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- imports and variables
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
local ScenarioUtils = import('/lua/sim/ScenarioUtilities.lua')
local ScenarioFramework = import('/lua/ScenarioFramework.lua')
local Utilities = import('/lua/utilities.lua')
local Entity = import('/lua/sim/Entity.lua').Entity
local Cinematics = import('/lua/cinematics.lua')

-- GENERAL GAME PLAY
local gatesDead = false -- true when all gates or defense object are destroyed
local unitsDead = false
local numEnemyUnits = {{}, {}, {}, {}, {}, {}, {}}
local killCount = {ARMY_1 = 0, ARMY_2 = 0, ARMY_3 = 0, ARMY_4 = 0, ARMY_SURVIVAL_ENEMY = 0, ARMY_SURVIVAL_ENEMY_SPECIALS = 0}
local nukeSub = {}

-- WAVE SPAWNING
local playerCount = 0 -- number of players (AI and human, but not including the enemy)
local waveCount = 0
local spawnAngle = 0

-- T4 SPAWNING
local spawnT4 = {nil, nil, nil, nil, nil, nil, nil, nil}

-- SUPER WEAPONS
local SWActive = false
local SWs = {nil, nil, nil, nil}

-- SPECIALS
local specialCount = 0
local numNukes = 1
local numUberNukes = 1
local specials = {{}, {}, {}, {}}
local special1Rate = 1
local special2Rate = 1
local special3Rate = 1

-- GATES
local numGatesOpen = 8 	-- how many spawn gates are alive
local gates = {nil, nil, nil, nil, nil, nil, nil, nil}

-- ARTILLERY
local numArtilleryOpen = 8 	-- how many Artillery are alive
local artillery = {nil, nil, nil, nil, nil, nil, nil, nil}

-- DIALOGS
--local optionsDialog = { Army_1 = optionsDialogArmy1(), Army_2 = optionsDialogArmy2(), Army_3 = optionsDialogArmy3(), Army_4 = optionsDialogArmy4() }
local allPlayerDialogs = {{}, {}, {}, {}, {}}
local dialogSettings = {{3, 15, true}, {3, 15, true}, {3, 15, true}, {3, 15, true}, {3, 15, false}}

-- TAUNTS
local usedTaunts = {}

-- MISSION


------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- prepare unit tables
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
local landUnits = { Tech1 = {Shield = {}, AntiAir = {}, DirectFire = {} }, Tech2 = {Shield = {}, AntiAir = {}, DirectFire = {} }, Tech3 = {Shield = {}, AntiAir = {}, DirectFire = {}, Special = { 'ual0301', 'url0301', 'uel0301', 'xsl0301', 'urs0201', 'urs0201', 'urs0201', 'urs0201', } }, Tech4 = {} }
local airUnits = { Tech1 = {Transport = {}, Carrier = {}, AntiAir = {}, DirectFire = {} }, Tech2 = {Transport = {}, Carrier = {}, AntiAir = {}, DirectFire = {} }, Tech3 = {Transport = {}, Carrier = {}, AntiAir = {}, DirectFire = {} }, Tech4 = {} }
local specialUnits = {{'xeb2402'}, {'xab2307'}, {'xsb2305', 'ueb2305', 'urb2305', 'uab2305'}, {'xsb2401'}, {'uab2302', 'urb2302', 'ueb2302', 'xsb2302'}}

local specialCondition = {
	'Last Artillery destroyed',
	'Defense objects activated',
	'First gate destroyed',
	'Last gate destroyed',
}

-- upgrade list for SCU's
local upgradeTable = {
	-- aeon
	A={
		'Shield',
		'ShieldHeavy',
		'StabilitySuppressant',
		'SystemIntegrityCompensator',
		},
	-- cybran
	R={
		'CloakingGenerator',
		'EMPCharge',
		'FocusConvertor',
		'NaniteMissileSystem',
		'SelfRepairSystem',
		'StealthGenerator', -- NOTE requires cloaking generator first
		},
	-- uef
	E={
		'AdvancedCoolingUpgrade',
		'HighExplosiveOrdnance',
		'RadarJammer',
		'Shield',
		'ShieldGeneratorField',
		},
	-- seraphim
	S={
		'DamageStabilization',
		'Missile',
		'Overcharge',
		'Shield',
		},
}

-- defines possible formations for the units
local formationTable = {
	'AttackFormation',
	'GrowthFormation',
	'SixWideFormation',
	'TravellingFormation',
}

------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- engine invoked main functions
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

-- this function is called once before game start, it does some player initialization
function OnPopulate()
	-- make default settings work
	for i = 1,table.getn(import('/maps/5iver Survival Land.v0001/5iver Survival Land_options.lua').options) do
		if (ScenarioInfo.Options[import('/maps/5iver Survival Land.v0001/5iver Survival Land_options.lua').options[i].key] == nil) then
			ScenarioInfo.Options[import('/maps/5iver Survival Land.v0001/5iver Survival Land_options.lua').options[i].key] = import('/maps/5iver Survival Land.v0001/5iver Survival Land_options.lua').options[i].values[1].key
		end
	end

	-- setup prebuilt units
	if (ScenarioInfo.Options.PrebuiltUnits == 'On') then
		ScenarioInfo.Options.PrebuiltUnits = 'Off'
		ScenarioInfo.Options.opt_General_PrebuiltUnits = 1
	else
		ScenarioInfo.Options.opt_General_PrebuiltUnits = 0
	end

	if (ScenarioInfo.Options.opt_General_PeakTechTime == 0) then
		ScenarioInfo.Options.opt_General_PeakTechTime = ScenarioInfo.Options.opt_Defense_ObjectStartup
	end

	-- modify options if Presets are used
	if (ScenarioInfo.Options.opt_Auto_Preset == 8) then
		ScenarioInfo.Options.opt_General_Difficulty = 10
		ScenarioInfo.Options.opt_Defense_GateDeathT4s = 8
		ScenarioInfo.Options.opt_Defense_Special = 5
		ScenarioInfo.Options.opt_Defense_Artillery = 5
		ScenarioInfo.Options.opt_Defense_Gate = 5
		ScenarioInfo.Options.opt_Defense_ObjectRegen = 5
		ScenarioInfo.Options.opt_General_StartupDelay = 0
		ScenarioInfo.Options.opt_HealthBuffer_Land = 5
		ScenarioInfo.Options.opt_HealthBuffer_Air = 5
		ScenarioInfo.Options.opt_General_WavePeriod = 1
		ScenarioInfo.Options.opt_Defense_ObjectStartup = 25
		ScenarioInfo.Options.opt_NumUnits_Land = 50
		ScenarioInfo.Options.opt_NumUnits_Air = 60
		ScenarioInfo.Options.opt_General_PeakTechTime = 30
		ScenarioInfo.Options.opt_General_UnitCap = 1500
		ScenarioInfo.Options.opt_Veterancy_Land = 5
		ScenarioInfo.Options.opt_Veterancy_Air = 5
	elseif (ScenarioInfo.Options.opt_Auto_Preset == 7) then
		ScenarioInfo.Options.opt_General_Difficulty = 10
		ScenarioInfo.Options.opt_Defense_GateDeathT4s = 8
		ScenarioInfo.Options.opt_Defense_Special = 5
		ScenarioInfo.Options.opt_Defense_Artillery = 5
		ScenarioInfo.Options.opt_Defense_Gate = 5
		ScenarioInfo.Options.opt_Defense_ObjectRegen = 5
		ScenarioInfo.Options.opt_General_StartupDelay = 0
		ScenarioInfo.Options.opt_HealthBuffer_Land = 10
		ScenarioInfo.Options.opt_HealthBuffer_Air = 10
		ScenarioInfo.Options.opt_General_WavePeriod = 1
		ScenarioInfo.Options.opt_Defense_ObjectStartup = 30
		ScenarioInfo.Options.opt_NumUnits_Land = 20
		ScenarioInfo.Options.opt_NumUnits_Air = 25
		ScenarioInfo.Options.opt_General_PeakTechTime = 35
		ScenarioInfo.Options.opt_General_UnitCap = 1500
		ScenarioInfo.Options.opt_Veterancy_Land = 5
		ScenarioInfo.Options.opt_Veterancy_Air = 5
	elseif (ScenarioInfo.Options.opt_Auto_Preset == 6) then
		ScenarioInfo.Options.opt_General_Difficulty = 10
		ScenarioInfo.Options.opt_Defense_GateDeathT4s = 8
		ScenarioInfo.Options.opt_Defense_Special = 5
		ScenarioInfo.Options.opt_Defense_Artillery = 5
		ScenarioInfo.Options.opt_Defense_Gate = 5
		ScenarioInfo.Options.opt_Defense_ObjectRegen = 5
		ScenarioInfo.Options.opt_General_StartupDelay = 0
		ScenarioInfo.Options.opt_HealthBuffer_Land = 2
		ScenarioInfo.Options.opt_HealthBuffer_Air = 2
		ScenarioInfo.Options.opt_General_WavePeriod = 1
		ScenarioInfo.Options.opt_Defense_ObjectStartup = 30
		ScenarioInfo.Options.opt_NumUnits_Land = 100
		ScenarioInfo.Options.opt_NumUnits_Air = 125
		ScenarioInfo.Options.opt_General_PeakTechTime = 35
		ScenarioInfo.Options.opt_General_UnitCap = 1500
		ScenarioInfo.Options.opt_Veterancy_Land = 5
		ScenarioInfo.Options.opt_Veterancy_Air = 5
	elseif (ScenarioInfo.Options.opt_Auto_Preset == 5) then
		ScenarioInfo.Options.opt_General_Difficulty = 10
		ScenarioInfo.Options.opt_Defense_GateDeathT4s = 8
		ScenarioInfo.Options.opt_Defense_Special = 5
		ScenarioInfo.Options.opt_Defense_Artillery = 5
		ScenarioInfo.Options.opt_Defense_Gate = 5
		ScenarioInfo.Options.opt_Defense_ObjectRegen = 5
		ScenarioInfo.Options.opt_General_StartupDelay = 0
		ScenarioInfo.Options.opt_HealthBuffer_Land = 5
		ScenarioInfo.Options.opt_HealthBuffer_Air = 5
		ScenarioInfo.Options.opt_General_WavePeriod = 1
		ScenarioInfo.Options.opt_Defense_ObjectStartup = 30
		ScenarioInfo.Options.opt_NumUnits_Land = 40
		ScenarioInfo.Options.opt_NumUnits_Air = 50
		ScenarioInfo.Options.opt_General_PeakTechTime = 35
		ScenarioInfo.Options.opt_General_UnitCap = 1500
		ScenarioInfo.Options.opt_Veterancy_Land = 5
		ScenarioInfo.Options.opt_Veterancy_Air = 5
	elseif (ScenarioInfo.Options.opt_Auto_Preset == 4) then
		ScenarioInfo.Options.opt_General_Difficulty = 10
		ScenarioInfo.Options.opt_Defense_GateDeathT4s = 8
		ScenarioInfo.Options.opt_Defense_Special = 5
		ScenarioInfo.Options.opt_Defense_Artillery = 5
		ScenarioInfo.Options.opt_Defense_Gate = 5
		ScenarioInfo.Options.opt_Defense_ObjectRegen = 5
		ScenarioInfo.Options.opt_General_StartupDelay = 0
		ScenarioInfo.Options.opt_HealthBuffer_Land = 5
		ScenarioInfo.Options.opt_HealthBuffer_Air = 5
		ScenarioInfo.Options.opt_General_WavePeriod = 5/60
		ScenarioInfo.Options.opt_Defense_ObjectStartup = 30
		ScenarioInfo.Options.opt_NumUnits_Land = 40
		ScenarioInfo.Options.opt_NumUnits_Air = 50
		ScenarioInfo.Options.opt_General_PeakTechTime = 35
		ScenarioInfo.Options.opt_General_UnitCap = 1500
		ScenarioInfo.Options.opt_Veterancy_Land = 5
		ScenarioInfo.Options.opt_Veterancy_Air = 5
	elseif (ScenarioInfo.Options.opt_Auto_Preset == 3) then
		ScenarioInfo.Options.opt_General_Difficulty = 5
		ScenarioInfo.Options.opt_Defense_GateDeathT4s = 9
		ScenarioInfo.Options.opt_Defense_Special = 0
		ScenarioInfo.Options.opt_Defense_Artillery = 1
		ScenarioInfo.Options.opt_Defense_Gate = 3
		ScenarioInfo.Options.opt_Defense_ObjectRegen = 5
		ScenarioInfo.Options.opt_General_StartupDelay = 5
		ScenarioInfo.Options.opt_HealthBuffer_Land = 2.5
		ScenarioInfo.Options.opt_HealthBuffer_Air = 2.5
		ScenarioInfo.Options.opt_General_WavePeriod = 1
		ScenarioInfo.Options.opt_Defense_ObjectStartup = 30
		ScenarioInfo.Options.opt_NumUnits_Land = 30
		ScenarioInfo.Options.opt_NumUnits_Air = 38
		ScenarioInfo.Options.opt_General_PeakTechTime = 45
		ScenarioInfo.Options.opt_General_UnitCap = 1500
		ScenarioInfo.Options.opt_Veterancy_Land = 2
		ScenarioInfo.Options.opt_Veterancy_Air = 2
	elseif (ScenarioInfo.Options.opt_Auto_Preset == 2) then
		ScenarioInfo.Options.opt_General_Difficulty = 5
		ScenarioInfo.Options.opt_Defense_GateDeathT4s = 9
		ScenarioInfo.Options.opt_Defense_Special = 0
		ScenarioInfo.Options.opt_Defense_Artillery = 1
		ScenarioInfo.Options.opt_Defense_Gate = 3
		ScenarioInfo.Options.opt_Defense_ObjectRegen = 5
		ScenarioInfo.Options.opt_General_StartupDelay = 5
		ScenarioInfo.Options.opt_HealthBuffer_Land = 2.5
		ScenarioInfo.Options.opt_HealthBuffer_Air = 2.5
		ScenarioInfo.Options.opt_General_WavePeriod = 5/60
		ScenarioInfo.Options.opt_Defense_ObjectStartup = 30
		ScenarioInfo.Options.opt_NumUnits_Land = 30
		ScenarioInfo.Options.opt_NumUnits_Air = 38
		ScenarioInfo.Options.opt_General_PeakTechTime = 45
		ScenarioInfo.Options.opt_General_UnitCap = 1500
		ScenarioInfo.Options.opt_Veterancy_Land = 2
		ScenarioInfo.Options.opt_Veterancy_Air = 2
	elseif (ScenarioInfo.Options.opt_Auto_Preset == 1) then
		ScenarioInfo.Options.opt_General_Difficulty = 2
		ScenarioInfo.Options.opt_Defense_GateDeathT4s = 10
		ScenarioInfo.Options.opt_Defense_Special = 0
		ScenarioInfo.Options.opt_Defense_Artillery = 0
		ScenarioInfo.Options.opt_Defense_Gate = 1
		ScenarioInfo.Options.opt_Defense_ObjectRegen = 2
		ScenarioInfo.Options.opt_General_StartupDelay = 10
		ScenarioInfo.Options.opt_HealthBuffer_Land = 1
		ScenarioInfo.Options.opt_HealthBuffer_Air = 1
		ScenarioInfo.Options.opt_General_WavePeriod = 1
		ScenarioInfo.Options.opt_Defense_ObjectStartup = 25
		ScenarioInfo.Options.opt_NumUnits_Land = 20
		ScenarioInfo.Options.opt_NumUnits_Air = 25
		ScenarioInfo.Options.opt_General_PeakTechTime = 55
		ScenarioInfo.Options.opt_General_UnitCap = 1500
		ScenarioInfo.Options.opt_Veterancy_Land = 0
		ScenarioInfo.Options.opt_Veterancy_Air = 0
	end

	ScenarioUtils.InitializeArmies()
	ScenarioFramework.SetPlayableArea('AREA_1', false)

	-- count number of players
	local armies = ListArmies()
	playerCount = table.getn(armies) - 2	-- number of armies minus one ai player
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": playerCount: "..playerCount)

	--allTech1 = EntityCategoryGetUnitList(categories.TECH1)
	--local cat = categories.ALLUNITS - categories.URL0402

	-- build unit table from all available units (including enabled mods)
	for id,bp in pairs(__blueprints) do
		if (bp.Categories and string.len(id) == 7) then
			if table.find(bp.Categories,'MOBILE') and not table.find(bp.Categories,'ENGINEER') and not table.find(bp.Categories,'SCOUT') and not table.find(bp.Categories,'SATELLITE') and not table.find(bp.Categories,'INSIGNIFICANTUNIT') and not table.find(bp.Categories,'TACTICALMISSILEPLATFORM') then --and not table.find(bp.Categories,'NUKE')then
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": id: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
				if table.find(bp.Categories,'LAND') and not table.find(bp.Categories,'INTELLIGENCE') then
					if table.find(bp.Categories,'TECH1') and table.find(bp.Categories,'BUILTBYTIER3FACTORY') then
						if table.find(bp.Categories,'SHIELD') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech1 Land Shield: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(landUnits.Tech1.Shield, id)
						elseif table.find(bp.Categories,'ANTIAIR') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech1 Land AntiAir: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(landUnits.Tech1.AntiAir, id)
						else
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech1 Land DirectFire: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(landUnits.Tech1.DirectFire, id)
						end
					elseif table.find(bp.Categories,'TECH2') and table.find(bp.Categories,'BUILTBYTIER3FACTORY') and (id ~= "xrl0302") then
						if table.find(bp.Categories,'SHIELD') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech2 Land Shield: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(landUnits.Tech2.Shield, id)
						elseif table.find(bp.Categories,'ANTIAIR') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech2 Land AntiAir: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(landUnits.Tech2.AntiAir, id)
						else
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech2 Land DirectFire: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(landUnits.Tech2.DirectFire, id)
						end
					elseif table.find(bp.Categories,'TECH3') and table.find(bp.Categories,'BUILTBYTIER3FACTORY') then
						if table.find(bp.Categories,'SHIELD') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech3 Land Shield: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(landUnits.Tech3.Shield, id)
						elseif table.find(bp.Categories,'ANTIAIR') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech3 Land AntiAir: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(landUnits.Tech3.AntiAir, id)
						else
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech3 Land DirectFire: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(landUnits.Tech3.DirectFire, id)
						end
					elseif table.find(bp.Categories,'EXPERIMENTAL') and (id ~= "url0401") and (id ~= "xsl0007") then --and (id ~= "xel0401") then Scathis, piece of Seraphim Assault Bot, UEF Goliath
						--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech4 Land: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
						table.insert(landUnits.Tech4, id)
					end
				elseif table.find(bp.Categories,'AIR') and (id ~= "daa0206") and (id ~= "uaa0302") and (id ~= "ura0302") and (id ~= "uea0302") and (id ~= "xsa0302") then --not guided missiles or T3 spy planes
					if table.find(bp.Categories,'TECH1') and (table.find(bp.Categories,'BUILTBYTIER3FACTORY') or table.find(bp.Categories,'TRANSPORTBUILTBYTIER3FACTORY')) then
						if table.find(bp.Categories,'TRANSPORTATION') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech1 Air Transportation: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech1.Transport, id)
						elseif table.find(bp.Categories,'CARRIER') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech1 Air Carrier: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech1.Carrier, id)
						elseif table.find(bp.Categories,'BOMBER') or table.find(bp.Categories,'GROUNDATTACK') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech1 Air DirectFire: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech1.DirectFire, id)
						else
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech1 Air AntiAir: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech1.AntiAir, id)
						end
					elseif table.find(bp.Categories,'TECH2') and (table.find(bp.Categories,'BUILTBYTIER3FACTORY') or table.find(bp.Categories,'TRANSPORTBUILTBYTIER3FACTORY')) then
						if table.find(bp.Categories,'TRANSPORTATION') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech2 Air Transportation: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech2.Transport, id)
						elseif table.find(bp.Categories,'CARRIER') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech2 Air Carrier: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech2.Carrier, id)
						elseif table.find(bp.Categories,'BOMBER') or table.find(bp.Categories,'GROUNDATTACK') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech2 Air DirectFire: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech2.DirectFire, id)
						else
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech2 Air AntiAir: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech2.AntiAir, id)
						end
					elseif table.find(bp.Categories,'TECH3') and (table.find(bp.Categories,'BUILTBYTIER3FACTORY') or table.find(bp.Categories,'TRANSPORTBUILTBYTIER3FACTORY')) then
						if table.find(bp.Categories,'TRANSPORTATION') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech3 Air Transportation: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech3.Transport, id)
						elseif table.find(bp.Categories,'CARRIER') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech3 Air Carrier: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech3.Carrier, id)
						elseif table.find(bp.Categories,'BOMBER') or table.find(bp.Categories,'GROUNDATTACK') then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech3 Air DirectFire: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech3.DirectFire, id)
						else
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech3 Air AntiAir: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
							table.insert(airUnits.Tech3.AntiAir, id)
						end
					elseif table.find(bp.Categories,'EXPERIMENTAL') then
						--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Tech4 Air: "..id.."   Faction: "..bp.General.FactionName.."   Name: "..bp.Description)
						table.insert(airUnits.Tech4, id)
					end
				end
			end
		end
	end
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Unit Table Built")
end

-- this function creates mex and hydrocarbon markers in save.lua
function ScenarioUtils.CreateResources()
	-- fetch markers and iterate them
	local markers = ScenarioUtils.GetMarkers()
	for i, tblData in pairs(markers) do
		-- spawn resources?
		local doit = false
		if (tblData.resource and not tblData.SpawnWithArmy) then
			-- standard resources, spawn it
			doit = true
		elseif (tblData.resource and tblData.SpawnWithArmy) then
			-- resources bound to player, check if army is present
			for j, army in ListArmies() do
				if (tblData.SpawnWithArmy == army) then
					doit = true  -- we made sure the army is present, allow spawn
					break
				end
			end
		end

		if (doit) then
			-- check type of resource and set parameters
			local bp, albedo, sx, sz, lod
			if (tblData.type == "Mass") then
				albedo = "/env/common/splats/mass_marker.dds"
				bp = "/env/common/props/massDeposit01_prop.bp"
				sx = 2
				sz = 2
				lod = 100
			else
				albedo = "/env/common/splats/hydrocarbon_marker.dds"
				bp = "/env/common/props/hydrocarbonDeposit01_prop.bp"
				sx = 6
				sz = 6
				lod = 200
			end
			-- create the resource
			CreateResourceDeposit(tblData.type,	tblData.position[1], tblData.position[2], tblData.position[3], tblData.size)
			-- create the resource graphic on the map
			CreatePropHPR(bp, tblData.position[1], tblData.position[2], tblData.position[3], Random(0,360), 0, 0)
			-- create the resource icon on the map
			CreateSplat(
				tblData.position,           --# Position
				0,                          --# Heading (rotation)
				albedo,                     --# Texture name for albedo
				sx, sz,                     --# SizeX/Z
				lod,                        --# LOD
				0,                          --# Duration (0 == does not expire)
				-1,                         --# army (-1 == not owned by any single army)
				0							--# ???
			)
		end
	end
end

-- start of map script
function OnStart(self)
	-- gather game options
	--optionsDialog()

	-- initiate game modes and spawn players
	setupGame()

	-- start the game logic
	startScenarioThreads()

	-- set score calculation to show total kills
	for index, brain in ArmyBrains do
		brain.CalculateScore = function(thisBrain)
			return armyKills(thisBrain) + killCount[ArmyIndexer(thisBrain)]
		end
	end
end

setupGame = function()
	-- misc settings
	ScenarioInfo.Options.Victory = 'sandbox'	-- custom victory condition
	Utilities.UserConRequest("ui_ForceLifbarsOnEnemy")	--show enemy life bars

	-- setup army specific things
	for i, army in ListArmies() do
		if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4") then
			-- restrict players from building walls
			ScenarioFramework.AddRestriction(army, categories.WALL)
			-- set alliances
			SetAlliance(army, "ARMY_SURVIVAL_ENEMY", "Enemy")
			SetAlliance(army, "ARMY_SURVIVAL_ENEMY_SPECIALS", "Enemy")
			--SetAlliance(army, "ARMY_SURVIVAL_GATES", "Enemy")
			SetAlliedVictory(army, true)
			-- set alliance with all other hill players!
			for j, army2 in ListArmies() do
				if (army2 == "ARMY_1" or army2 == "ARMY_2" or army2 == "ARMY_3" or army2 == "ARMY_4") then
					SetAlliance(army, army2, "Ally")
				end
			end
			if (ScenarioInfo.Options.opt_General_UnitCap ~= 0) then
				SetArmyUnitCap(army, ScenarioInfo.Options.opt_General_UnitCap)
			end
		end
	end

	SetAlliance('ARMY_SURVIVAL_ENEMY', 'ARMY_SURVIVAL_ENEMY_SPECIALS', "Ally")

	SetArmyColor("ARMY_SURVIVAL_ENEMY", 255, 255, 255)
	SetArmyColor("ARMY_SURVIVAL_ENEMY_SPECIALS", 255, 20, 147)

	-- spawnwaves don't have a unit cap
	SetIgnoreArmyUnitCap("ARMY_SURVIVAL_ENEMY", true)

	spawnGates()
	spawnArtillery()
	spawnDefenseObjects()
	spawnPrebuiltEngineers()
	spawnPrebuiltUnits()
end

--creates the different scenario threads
startScenarioThreads = function()
	--start main thread
	local mainThread = ForkThread(function(self)
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Main thread running!")

		while (unitsDead == false) do
			WaitSeconds(1)
			gameLogic()
		end
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Killing mainThread")
		KillThread(self)
    end)

	-- start dialog thread
	local dialogThread = ForkThread(dialogThread)

	--this creates a new spawn thread only in survival mode
	local spawnThread = ForkThread(spawnThread)
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": spawnThread running!")

end

gameLogic = function()
	spawnGateT4s()
	if (numGatesOpen < 1) then -- check if all gates were destroyed
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": gamelogic() all gates destroyed")
		gatesDead = true
		local finalMessage = nil
		if (ScenarioInfo.Options.opt_General_EndGameScenario == 1) then
			if (unitsDead == false) then
				-- see if there are any enemy units left
				--local cats = categories.MOBILE + categories.NUKE + categories.UAB2302 + categories.URB2302 + categories.UEB2302 + categories.XSB2302 + categories.XAB2307 + categories.URL0401 + categories.UEB2401
				--local cat5 = categories.MOBILE + categories.NUKE + categories.UAB5555

				sumEnemyUnits, sumEnemyUnitsPrevious = unitsRemaining()

				if (sumEnemyUnits < 1) then
					unitsDead = true
					finalMessage = "All gates, Artillery and units destroyed"
				else
					if (sumEnemyUnits >= 100) then
						if (100 * math.ceil(sumEnemyUnits / 100) ~= 100 * math.ceil(sumEnemyUnitsPrevious / 100)) then
							--msgOut("Enemy units: <"..(100 * math.ceil(sumEnemyUnits / 100)), 1)
							msgOut("Enemy units remaining: "..sumEnemyUnits.."\nLand: "..numEnemyUnits[1].."\nAir: "..numEnemyUnits[2].."\nNaval: "..numEnemyUnits[3].."\nT4: "..numEnemyUnits[4], 2)
						end
					elseif (sumEnemyUnits < 100 and sumEnemyUnits >= 25) then
						if (25 * math.ceil(sumEnemyUnits / 25) ~= 25 * math.ceil(sumEnemyUnitsPrevious / 25)) then
							--msgOut("Enemy units: <"..(25 * math.ceil(sumEnemyUnits / 25)), 1)
							msgOut("Enemy units remaining: "..sumEnemyUnits.."\nLand: "..numEnemyUnits[1].."\nAir: "..numEnemyUnits[2].."\nNaval: "..numEnemyUnits[3].."\nT4: "..numEnemyUnits[4], 2)
						end
					elseif (sumEnemyUnits < 25 and sumEnemyUnits >= 5) then
						if (5 * math.ceil(sumEnemyUnits / 5) ~= 5 * math.ceil(sumEnemyUnitsPrevious / 5)) then
							--msgOut("Enemy units: <"..(5 * math.ceil(sumEnemyUnits / 5)), 1)
							msgOut("Enemy units remaining: "..sumEnemyUnits.."\nLand: "..numEnemyUnits[1].."\nAir: "..numEnemyUnits[2].."\nNaval: "..numEnemyUnits[3].."\nT4: "..numEnemyUnits[4], 2)
						end
					else
						if (sumEnemyUnits ~= sumEnemyUnitsPrevious) then
							--msgOut("Enemy units: "..sumEnemyUnits, 1)
							msgOut("Enemy units remaining: "..sumEnemyUnits.."\nLand: "..numEnemyUnits[1].."\nAir: "..numEnemyUnits[2].."\nNaval: "..numEnemyUnits[3].."\nT4: "..numEnemyUnits[4], 2)
						end
					end
				end
			end
		else
			unitsDead = true
			finalMessage = "All gates are destroyed"
			--GetArmyBrain("ARMY_SURVIVAL_ENEMY"):OnDefeat()
		end
		if (unitsDead == true) then
			if (ScenarioInfo.Options.opt_General_Taunts == 1) then
				ScenarioFramework.Dialogue(import('/maps/X1CA_004/X1CA_004_strings.lua').X04_M03_210, nil, true) -- took care of business
			end

			if (ScenarioInfo.Options.opt_Defense_Object < 6) then
				finalMessage = finalMessage.."\n\nThe defense objects have survived"
			end
			finalMessage = finalMessage.."\n\nYOU HAVE WON!"

			for i, army in ListArmies() do
				if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4") then
					GetArmyBrain(army):OnVictory()
				end
			end
			GetArmyBrain("ARMY_SURVIVAL_ENEMY"):OnDefeat()

			local displayEndGame = CreateDialogue(finalMessage, nil, 'center')

			--[[
			local focusArmy = GetFocusArmy()
			for i = 1,4 do
				if (focusArmy == i) then
					local displayEndGameThread = ForkThread(function(self)
						local displayEndGame = CreateDialogue(finalMessage.."\n\n\nSTARTUP OPTIONS:\n\n"..startupOptions().."\n\n", {'Close'}, 'center')
						displayEndGame.OnButtonPressed = function(self, button)
							if (button.buttonID == 1) then
								--SimConExecute("ui_lua RestartSession()")
								self:Destroy()
							end
						end
						KillThread(self)
					end)
				end
			end]]--

			EndGame()
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": ENDGAME - Victory.")
		end
	end
end

-- spawn 8 gates at the outer edges for enemy units to teleport in from
spawnGates = function()
	numGatesOpen = 8
	for i = 1,8 do
		local gateName = "GATE"..i
		local pos = ScenarioUtils.MarkerToPosition(gateName)
		local r = Utilities.GetRandomInt(1,4)
		local gateFaction
		if (r == 1) then
			gateFaction = 'UAB0304'
		elseif (r == 2) then
			gateFaction = 'URB0304'
		elseif (r == 3) then
			gateFaction = 'UEB0304'
		else
			gateFaction = 'XSB0304'
		end
		gates[i] = CreateUnitHPR( gateFaction, "ARMY_SURVIVAL_ENEMY", pos[1], pos[2], pos[3], 0,0,0)
		gates[i]:SetMaxHealth(110000 * ScenarioInfo.Options.opt_General_Difficulty)
		gates[i]:SetHealth(nil, (110000 * ScenarioInfo.Options.opt_General_Difficulty))
		gates[i]:SetReclaimable(false)
		gates[i]:SetCapturable(false)
		gates[i]:SetProductionPerSecondEnergy(9999)
		gates[i]:SetProductionPerSecondMass(9999)
		gates[i]:SetCustomName("Gate "..i)

		-- remove gate wreckage
		local bp = gates[i]:GetBlueprint()
		if (bp ~= nil) then
			bp.Wreckage = nil -- bp.Wreckage.WreckageLayers.Land=false
		end
		spawnDefenses(pos, ScenarioInfo.Options.opt_Defense_Gate, "ARMY_SURVIVAL_ENEMY")

		-- set gate onkilled function
		gates[i].OldOnKilled = gates[i].OnKilled
		gates[i].myID = i
		gates[i].OnKilled = function(self, instigator, type, overkillRatio)

			--spawn a nuke
			nuke(self:GetPosition())

			-- add points to player
			local instArmy = instigator:GetArmy()
			local armyName = ArmyIndexer(GetArmyBrain(instArmy))
			local scoreMod = math.floor(1 - (GetGameTimeSeconds() / 3600))
			if scoreMod > 0 then
				killCount[armyName] = killCount[armyName] + (1000000 * scoreMod)
			end

			-- handle game logic
			numGatesOpen = numGatesOpen - 1
			msgOut(gateName.." destroyed by "..getUsername(instArmy).."\n\nGates remaining: "..numGatesOpen, 2)
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": "..gateName.." destroyed by "..getUsername(instArmy).."\n\n"..numGatesOpen.." gates remaining")

			if (numGatesOpen == 7) then
				spawnSpecials(3)
			elseif (numGatesOpen == 0) then
				spawnSpecials(4)
			end
			if (ScenarioInfo.Options.opt_General_Taunts == 1) then
				if (numGatesOpen == 0) then
					ScenarioFramework.Dialogue(import('/maps/X1CA_003/X1CA_003_strings.lua').X03_M02_082, nil, true) -- can't believe it
				else
					ScenarioFramework.Dialogue(import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M01_130, nil, true) -- scratch one
				end
			end
			spawnT4[self.myID] = 1
			self.myID = nil
			self.OldOnKilled(self, instigator, type, overkillRatio)
		end
	end
end

spawnArtillery = function()
	if (ScenarioInfo.Options.opt_Defense_Artillery > 0) then
		for i = 1,8 do
			local artyName = "ARTILLERY"..i
			local pos = ScenarioUtils.MarkerToPosition(artyName)
			local rangeBuff = 165 + (ScenarioInfo.Options.opt_General_Difficulty * 10)
			--artillery[i] = CreateUnitHPR( 'xab2307', "ARMY_SURVIVAL_ENEMY", pos[1], pos[2], pos[3], 0, math.rad(202.5 - 1 - (45 * i)), 0) -- subtract one or they don't rotate in the right direction
			artillery[i] = CreateUnitHPR( 'xab2307', "ARMY_SURVIVAL_ENEMY", pos[1], pos[2], pos[3], 0, math.rad(292.5 - 1 - (45 * i)), 0) -- subtract one or they don't rotate in the right direction
			artillery[i]:SetReclaimable(false)
			artillery[i]:SetCapturable(false)
			artillery[i]:SetProductionPerSecondEnergy(9999)
			artillery[i]:SetProductionPerSecondMass(9999)
			artillery[i]:SetCustomName("Artillery "..i)
			artillery[i]:SetIntelRadius('Vision', (rangeBuff - 10))
			local wp = artillery[i]:GetWeapon(1)
			wp:ChangeMaxRadius(rangeBuff)
			wp:ChangeMinRadius(80)

			-- remove artillery wreckage issue
			local bp = artillery[i]:GetBlueprint()
			if (bp ~= nil) then
				bp.Wreckage = nil -- bp.Wreckage.WreckageLayers.Land=false
			end

			spawnDefenses(pos, ScenarioInfo.Options.opt_Defense_Artillery, "ARMY_SURVIVAL_ENEMY")

			-- set onkilled function
			artillery[i].OldOnKilled = artillery[i].OnKilled
			artillery[i].myID = i
			artillery[i].OnKilled = function(self, instigator, type, overkillRatio)
				--spawn a nuke
				nuke(self:GetPosition())

				-- add points to player
				local instArmy = instigator:GetArmy()
				local armyName = ArmyIndexer(GetArmyBrain(instArmy))
				local scoreMod = math.floor(1 - (GetGameTimeSeconds() / 3600))
				if scoreMod > 0 then
					killCount[armyName] = killCount[armyName] + (100000 * scoreMod)
				end

				-- handle game logic
				numArtilleryOpen = numArtilleryOpen - 1
				if (numArtilleryOpen > 0) then
					msgOut(artyName.." destroyed by "..getUsername(instArmy).."\n\nArtillery remaining: "..numArtilleryOpen, 2)
				else
					msgOut(artyName.." destroyed by "..getUsername(instArmy))
				end

				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": "..artyName.." destroyed by "..getUsername(instArmy).."\n\n"..numArtilleryOpen.." Artillery remaining")
				self.myID = nil
				if (numArtilleryOpen == 0) then
					spawnSpecials(1)
				end
				if (ScenarioInfo.Options.opt_General_Taunts == 1) then
					if (numArtilleryOpen == 0) then
						ScenarioFramework.Dialogue(import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M01_200, nil, true) -- eigth
					elseif (numArtilleryOpen == 7) then
						ScenarioFramework.Dialogue(import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M01_160, nil, true) -- first
					elseif (numArtilleryOpen == 6) then
						ScenarioFramework.Dialogue(import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M01_170, nil, true) -- second
					elseif (numArtilleryOpen == 4) then
						ScenarioFramework.Dialogue(import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M01_190, nil, true) -- fourth
					elseif (numArtilleryOpen == 1) then
						ScenarioFramework.Dialogue(import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M01_180, nil, true) -- seventh
					else
						ScenarioFramework.Dialogue(import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M01_210, nil, true) -- some remaining
					end
				end
				self.OldOnKilled(self, instigator, type, overkillRatio)
			end
		end
		showArtilleryRange()
	else
		specialCount = specialCount + 1
		numArtilleryOpen = 0
	end
end

-- spawn the defense object
spawnDefenseObjects = function()
	--showSpecialDestroyArea()
	if (ScenarioInfo.Options.opt_Defense_Object < 6) then
		local object = nil
		if (ScenarioInfo.Options.opt_Defense_Object == 1) then	-- aeon exp. resource generator (default)
			object = 'XAB1401'
		elseif (ScenarioInfo.Options.opt_Defense_Object == 2) then	-- aeon rapid fire arty
			object = 'XAB2307'
		elseif (ScenarioInfo.Options.opt_Defense_Object == 3) then	-- scathis
			object = 'URL0401'
		elseif (ScenarioInfo.Options.opt_Defense_Object == 4) then	-- mavor
			object = 'UEB2401'
		elseif (ScenarioInfo.Options.opt_Defense_Object == 5) then	-- black sun
			object = 'UEC1902'
		end

		local pos = ScenarioUtils.MarkerToPosition("HILL_CENTER")
		for i, army in ListArmies() do	-- assign superweapons to each player
			if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4") then
				if (army == "ARMY_1") then
					SWs[i] = CreateUnitHPR( object, army, pos[1] - 3.6, pos[2], pos[3] - 3.6, 0,0,0)
				elseif (army == "ARMY_2") then
					SWs[i] = CreateUnitHPR( object, army, pos[1] + 3.6, pos[2], pos[3] - 3.6, 0,0,0)
				elseif (army == "ARMY_3") then
					SWs[i] = CreateUnitHPR( object, army, pos[1] + 3.6, pos[2], pos[3] + 3.6, 0,0,0)
				elseif (army == "ARMY_4") then
					SWs[i] = CreateUnitHPR( object, army, pos[1] - 3.6, pos[2], pos[3] + 3.6, 0,0,0)
				end

				SWs[i]:SetReclaimable(false)
				SWs[i]:SetCapturable(false)
				SWs[i]:SetImmobile(true)
				--SWs[i]:SetUnSelectable(true)
				SWs[i]:SetCustomName("Deactivated")

				local bp = gates[i]:GetBlueprint()
				table.insert(bp.Categories, 'SPECIALHIGHPRI')

				if (ScenarioInfo.Options.opt_Defense_ObjectRegen == 4) then
					SWs[i]:SetRegenRate(8.5) --about 500hp/min
				elseif (ScenarioInfo.Options.opt_Defense_ObjectRegen == 3) then
					SWs[i]:SetRegenRate(34) --about 2,000hp/min
				elseif (ScenarioInfo.Options.opt_Defense_ObjectRegen == 2) then
					SWs[i]:SetRegenRate(85) --about 5,000hp/min
				elseif (ScenarioInfo.Options.opt_Defense_ObjectRegen == 1) then
					SWs[i]:SetRegenRate(1020) --about 60,000hp/min
				elseif (ScenarioInfo.Options.opt_Defense_ObjectRegen == 0) then
					SWs[i]:SetCanTakeDamage(false)
					SWs[i]:SetCanBeKilled(false)
					SWs[i]:SetDoNotTarget(true)
				end

				if (ScenarioInfo.Options.opt_Defense_Object == 1) then
					SWs[i]:SetProductionActive(false)
				elseif (ScenarioInfo.Options.opt_Defense_Object == 5) then

				else
					local wp = SWs[i]:GetWeapon(1)
					wp:ChangeRateOfFire(2)
					wp:SetWeaponEnabled(false)
					--wp:ChangeMaxRadius(555) -- 300
					--wp:ChangeMinRadius(100) -- 300
					local wpBp = wp:GetBlueprint()
					wpBp.EnergyDrainPerSecond = 0	-- doesn't drain energy
					wpBp.EnergyRequired = 0 -- doesn't need energy
				end

				-- set onkilled function
				SWs[i].OldOnKilled = SWs[i].OnKilled
				SWs[i].OnKilled = function(self, instigator, type, overkillRatio)

					--spawn a nuke (removed so that the cause of the death could be seen)
					--nuke(self:GetPosition())
					if (ScenarioInfo.Options.opt_General_Taunts == 1) then
						ScenarioFramework.Dialogue(import('/maps/X1CA_005/X1CA_005_strings.lua').X05_DB01_030, nil, true) -- Op has ended in failure
					end

					if (unitsDead == false) then
						self.OldOnKilled(self, instigator, type, overkillRatio)
						for j, army in ListArmies() do
							if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4") then
								killAllCommanders(army)
								GetArmyBrain(army):OnDefeat()
							end
						end
						GetArmyBrain("ARMY_SURVIVAL_ENEMY"):OnVictory()

						--local displayEndGame = CreateDialogue("Defense object destroyed\n\nYOU HAVE LOST!\n\n\nSTARTUP OPTIONS:\n\n"..startupOptions().."\n\n", {'Close'}, 'center')
						local displayEndGame = CreateDialogue("Defense object destroyed\n\nYOU HAVE LOST!", nil, 'center')
						--[[displayEndGame.OnButtonPressed = function(self, button)
							if (button.buttonID == 1) then
								--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": focusArmy == "..focusArmy.."   button.presser = "..button.presser)
								self:Destroy()
								--SimConExecute("ui_lua RestartSession()")
							end
						end]]--

						--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": ENDGAME - Defense object destroyed.")
						EndGame()
					end
				end
			end
		end
	else
		specialCount = specialCount + 1
	end
end

-- spawn prebuilt engineers if the option was selected
spawnPrebuiltEngineers = function()
	if (ScenarioInfo.Options.opt_General_PrebuiltEngineers > 0) then
		local units = {{'ual0105', 'url0105', 'uel0105', 'xsl0105'}, {'ual0208', 'url0208', 'uel0208', 'xsl0208'}, {'ual0309', 'url0309', 'uel0309', 'xsl0309'}, {'ual0301', 'url0301', 'uel0301', 'xsl0301'}}
		local prebuiltEngineers = {}
		for i, army in ListArmies() do
			if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4") then
				for j = 1,4 do
					if (army == "ARMY_"..j) then
						pos = ScenarioUtils.MarkerToPosition('ARMY_'..j)
						prebuiltEngineers[j * 1] = CreateUnitHPR( units[ScenarioInfo.Options.opt_General_PrebuiltEngineers][1], army, pos[1] - 5, pos[2], pos[3], 0,0,0)
						prebuiltEngineers[j * 2] = CreateUnitHPR( units[ScenarioInfo.Options.opt_General_PrebuiltEngineers][2], army, pos[1] + 5, pos[2], pos[3], 0,0,0)
						prebuiltEngineers[j * 3] = CreateUnitHPR( units[ScenarioInfo.Options.opt_General_PrebuiltEngineers][3], army, pos[1] - 10, pos[2], pos[3], 0,0,0)
						prebuiltEngineers[j * 4] = CreateUnitHPR( units[ScenarioInfo.Options.opt_General_PrebuiltEngineers][4], army, pos[1] + 10, pos[2], pos[3], 0,0,0)
					end
				end
			end
		end
	end
end

-- spawn prebuilt units if the option was selected
spawnPrebuiltUnits = function()
	if (ScenarioInfo.Options.opt_General_PrebuiltUnits == 1) then
		for i, army in ListArmies() do
			if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4") then
				local factionIndex = GetArmyBrain(army):GetFactionIndex()
				local object
				local pos
				local prebuiltUnits = { Factory = {}, Hydrocarbon = {}, Mass = {}, Engineer = {}}

				if factionIndex == 1 then
					object1 = 'ueb0101'
					object2 = 'ueb1102'
					object3 = 'ueb1103'
				elseif factionIndex == 2 then
					object1 = 'uab0101'
					object2 = 'uab1102'
					object3 = 'uab1103'
				elseif factionIndex == 3 then
					object1 = 'urb0101'
					object2 = 'urb1102'
					object3 = 'urb1103'
				elseif factionIndex == 4 then
					object1 = 'xsb0101'
					object2 = 'xsb1102'
					object3 = 'xsb1103'
				end

				for j = 1,4 do
					if (army == "ARMY_"..j) then
						for k = 1,2 do
							pos = ScenarioUtils.MarkerToPosition('Factory'..j..'_'..k)
							prebuiltUnits.Factory[(2 * j) - k + 1] = CreateUnitHPR( object1, army, pos[1], pos[2], pos[3], 0,0,0)

							pos = ScenarioUtils.MarkerToPosition('HYDROCARBON'..j..'_'..k)
							prebuiltUnits.Hydrocarbon[(2 * j) - k + 1] = CreateUnitHPR( object2, army, pos[1], pos[2], pos[3], 0,0,0)
						end

						for k = 1,4 do
							pos = ScenarioUtils.MarkerToPosition('MASS'..j..'_'..k)
							prebuiltUnits.Mass[(4 * j) - k + 1] = CreateUnitHPR( object3, army, pos[1], pos[2], pos[3], 0,0,0)
						end
					end
				end
			end
		end
	end
end

spawnDefenses = function(pos, level, army)
	local defenseUnits = { Sam = { {}, {}, {}, {}, {}, {}, {}, {} }, PD = { {}, {}, {}, {}, {}, {}, {}, {} }, Shield = { {}, {}, {}, {} }, T2Arty = { {}, {}, {}, {}, {}, {}, {}, {} }}
	if (level > 1) then
		defenseUnits.Sam[1] = CreateUnitHPR( 'xsb2304', army, pos[1] + 5, pos[2], pos[3] + 5, 0,0,0)
		defenseUnits.Sam[2] = CreateUnitHPR( 'xsb2304', army, pos[1] + 5, pos[2], pos[3] - 5, 0,0,0)
		defenseUnits.Sam[3] = CreateUnitHPR( 'xsb2304', army, pos[1] - 5, pos[2], pos[3] + 5, 0,0,0)
		defenseUnits.Sam[4] = CreateUnitHPR( 'xsb2304', army, pos[1] - 5, pos[2], pos[3] - 5, 0,0,0)
		for i = 1,4 do
			--unit = CreateUnitHPR( 'xsb2304', army, pos[1] + 5 * cos(180 + 22.5 + 45 * i), pos[2], pos[3] + 5 * sin(i * math.pi / 2), 0,0,0)
			defenseUnits.Sam[i]:SetVeterancy(5)
			defenseUnits.Sam[i]:SetProductionPerSecondEnergy(9999)
			defenseUnits.Sam[i]:SetProductionPerSecondMass(9999)
		end
	end
	if (level > 2) then
		 defenseUnits.PD[1] = CreateUnitHPR( 'xeb2306', army, pos[1] + 10, pos[2], pos[3] + 10, 0,0,0)
		 defenseUnits.PD[2] = CreateUnitHPR( 'xeb2306', army, pos[1] + 10, pos[2], pos[3] - 10, 0,0,0)
		 defenseUnits.PD[3] = CreateUnitHPR( 'xeb2306', army, pos[1] - 10, pos[2], pos[3] + 10, 0,0,0)
		 defenseUnits.PD[4] = CreateUnitHPR( 'xeb2306', army, pos[1] - 10, pos[2], pos[3] - 10, 0,0,0)
		for i = 1,4 do
			defenseUnits.PD[i]:SetVeterancy(5)
			defenseUnits.PD[i]:SetProductionPerSecondEnergy(9999)
			defenseUnits.PD[i]:SetProductionPerSecondMass(9999)
		end
	end
	if (level > 3) then
		 defenseUnits.Sam[5] = CreateUnitHPR( 'xsb2304', army, pos[1] + 5, pos[2], pos[3], 0,0,0)
		 defenseUnits.Sam[6] = CreateUnitHPR( 'xsb2304', army, pos[1] - 5, pos[2], pos[3], 0,0,0)
		 defenseUnits.Sam[7] = CreateUnitHPR( 'xsb2304', army, pos[1], pos[2], pos[3] + 5, 0,0,0)
		 defenseUnits.Sam[8] = CreateUnitHPR( 'xsb2304', army, pos[1], pos[2], pos[3] - 5, 0,0,0)
		 defenseUnits.PD[5] = CreateUnitHPR( 'xeb2306', army, pos[1] + 10, pos[2], pos[3], 0,0,0)
		 defenseUnits.PD[6] = CreateUnitHPR( 'xeb2306', army, pos[1] - 10, pos[2], pos[3], 0,0,0)
		 defenseUnits.PD[7] = CreateUnitHPR( 'xeb2306', army, pos[1], pos[2], pos[3] + 10, 0,0,0)
		 defenseUnits.PD[8] = CreateUnitHPR( 'xeb2306', army, pos[1], pos[2], pos[3] - 10, 0,0,0)
		 defenseUnits.Shield[1] = CreateUnitHPR( 'xsb4301', army, pos[1] + 7.5, pos[2], pos[3] + 7.5, 0,0,0)
		 defenseUnits.Shield[2] = CreateUnitHPR( 'xsb4301', army, pos[1] + 7.5, pos[2], pos[3] - 7.5, 0,0,0)
		 defenseUnits.Shield[3] = CreateUnitHPR( 'xsb4301', army, pos[1] - 7.5, pos[2], pos[3] + 7.5, 0,0,0)
		 defenseUnits.Shield[4] = CreateUnitHPR( 'xsb4301', army, pos[1] - 7.5, pos[2], pos[3] - 7.5, 0,0,0)
		for i = 1,4 do
			defenseUnits.Sam[(i + 4)]:SetVeterancy(5)
			defenseUnits.PD[(i + 4)]:SetVeterancy(5)
			defenseUnits.Shield[i]:SetVeterancy(5)
			defenseUnits.Sam[i]:SetProductionPerSecondEnergy(9999)
			defenseUnits.Sam[i]:SetProductionPerSecondMass(9999)
			defenseUnits.PD[i]:SetProductionPerSecondEnergy(9999)
			defenseUnits.PD[i]:SetProductionPerSecondMass(9999)
			defenseUnits.Shield[i]:SetProductionPerSecondEnergy(9999)
			defenseUnits.Shield[i]:SetProductionPerSecondMass(9999)
		end
	end
	if (level > 4) then
		local AntiNuke1 = CreateUnitHPR( 'ueb4302', army, pos[1], pos[2], pos[3] - 7.5, 0,0,0)
		--AntiNuke1:SetIntelRadius('Vision', 180)
		AntiNuke1:GiveTacticalSiloAmmo(7)
		--AntiNuke1:SetAutoMode(true)
		--local wp = AntiNuke1:GetWeapon(1)
		--wp:ChangeMaxRadius(180)
		AntiNuke1:SetProductionPerSecondEnergy(9999)
		AntiNuke1:SetProductionPerSecondMass(9999)

		local AntiMissile1 = CreateUnitHPR( 'uab4201', army, pos[1], pos[2], pos[3] + 7.5, 0,0,0)
		--AntiMissile1:SetIntelRadius('Vision', 180)
		AntiMissile1:GiveTacticalSiloAmmo(7)
		--AntiNuke2:SetAutoMode(true)
		--local wp = AntiMissile1:GetWeapon(1)
		--wp:ChangeMaxRadius(180)
		AntiMissile1:SetProductionPerSecondEnergy(9999)
		AntiMissile1:SetProductionPerSecondMass(9999)

		local AntiMissile2 = CreateUnitHPR( 'urb4201', army, pos[1] - 7.5, pos[2], pos[3], 0,0,0)
		--AntiMissile2:SetIntelRadius('Vision', 180)
		AntiMissile2:GiveTacticalSiloAmmo(7)
		--AntiNuke2:SetAutoMode(true)
		--local wp = AntiMissile2:GetWeapon(1)
		--wp:ChangeMaxRadius(180)
		AntiMissile2:SetProductionPerSecondEnergy(9999)
		AntiMissile2:SetProductionPerSecondMass(9999)

		local AntiMissile3 = CreateUnitHPR( 'ueb4201', army, pos[1] + 7.5, pos[2], pos[3], 0,0,0)
		--AntiMissile3:SetIntelRadius('Vision', 180)
		AntiMissile3:GiveTacticalSiloAmmo(7)
		--AntiNuke2:SetAutoMode(true)
		--local wp = AntiMissile3:GetWeapon(1)
		--wp:ChangeMaxRadius(180)
		AntiMissile3:SetProductionPerSecondEnergy(9999)
		AntiMissile3:SetProductionPerSecondMass(9999)

		local adjust = 10 * math.tan(math.rad(22.5))
		--LOG("_____LOG: adjust = "..adjust)
		defenseUnits.T2Arty[1] = CreateUnitHPR( 'uab2303', army, pos[1] - adjust, pos[2], pos[3] - 10, 0,0,0)
		defenseUnits.T2Arty[2] = CreateUnitHPR( 'uab2303', army, pos[1] + adjust, pos[2], pos[3] - 10, 0,0,0)
		defenseUnits.T2Arty[3] = CreateUnitHPR( 'uab2303', army, pos[1] + 10, pos[2], pos[3] - adjust, 0,0,0)
		defenseUnits.T2Arty[4] = CreateUnitHPR( 'uab2303', army, pos[1] - 10, pos[2], pos[3] - adjust, 0,0,0)
		defenseUnits.T2Arty[5] = CreateUnitHPR( 'uab2303', army, pos[1] + 10, pos[2], pos[3] + adjust, 0,0,0)
		defenseUnits.T2Arty[6] = CreateUnitHPR( 'uab2303', army, pos[1] - 10, pos[2], pos[3] + adjust, 0,0,0)
		defenseUnits.T2Arty[7] = CreateUnitHPR( 'uab2303', army, pos[1] - adjust, pos[2], pos[3] + 10, 0,0,0)
		defenseUnits.T2Arty[8] = CreateUnitHPR( 'uab2303', army, pos[1] + adjust, pos[2], pos[3] + 10, 0,0,0)
		for i = 1,8 do
			defenseUnits.T2Arty[i]:SetVeterancy(5)
			defenseUnits.T2Arty[i]:SetProductionPerSecondEnergy(9999)
			defenseUnits.T2Arty[i]:SetProductionPerSecondMass(9999)
		end
	end
end

spawnSpecials = function(condition)
	msgOut(specialCondition[condition], 1)
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": "..specialCondition[condition])
	if (ScenarioInfo.Options.opt_Defense_Special > 0) then

		-- set variables
		specialCount = specialCount + 1
		local distance = math.random(1,4)
		local foundDistance = false

		-- find available distance
		while (foundDistance == false) do
			if (specials[distance][1] == nil and specials[distance][2] == nil and specials[distance][3] == nil and specials[distance][4] == nil) then
				foundDistance = true
			else
				distance = distance + 1
				if (distance > 4) then
					distance = 1
				end
			end
		end

		-- for each player, find an available angle and create a special
		for i = 1,playerCount do
			local angle = math.random(1,4)
			local foundAngle = false
			while (foundAngle == false) do
				if (specials[distance][angle] == nil) then

					-- destroy anything near the area where the special will spawn and then spawn the special
					local unit = specialUnits[specialCount][math.random(1,table.getn(specialUnits[specialCount]))]
					pos = {512.5 + (380 + 80 * distance) * math.cos(math.rad(135 + 90 * angle)), 0, 511.5 + (380 + 80 * distance) * math.sin(math.rad(135 + 90 * angle))}
					local unitsToDestroy = GetEntitiesInRect(pos[1] - 70, pos[3] - 70, pos[1] + 70, pos[3] + 70)
					if (unitsToDestroy ~= nil) then
						for i,v in ipairs(unitsToDestroy) do
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": GetArmyBrain(ARMY_SURVIVAL_ENEMY):GetArmyIndex(): "..GetArmyBrain("ARMY_SURVIVAL_ENEMY"):GetArmyIndex())
							if (v:GetArmy() ~= GetArmyBrain("ARMY_SURVIVAL_ENEMY"):GetArmyIndex() and v:GetArmy() ~= GetArmyBrain("ARMY_SURVIVAL_ENEMY_SPECIALS"):GetArmyIndex()) then
								v:Destroy()
							end
						end
					end

					specials[distance][angle] = CreateUnitHPR(unit, "ARMY_SURVIVAL_ENEMY_SPECIALS", pos[1], pos[2], pos[3], 0, math.rad(315 - 1 - (90 * angle)), 0) -- subtract one or they don't rotate the right direction
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": angle: "..angle.."   actual angle: "..(135 + 90 * angle).."   pitch: "..(315 - (90 * angle)))
					specials[distance][angle]:SetCustomName("Special "..distance.." - "..angle)
					specials[distance][angle]:SetProductionPerSecondEnergy(9999)
					specials[distance][angle]:SetProductionPerSecondMass(9999)
					specials[distance][angle]:SetVeterancy(5)
					specials[distance][angle]:SetMaxHealth(500 * specialCount * ScenarioInfo.Options.opt_General_Difficulty)
					specials[distance][angle]:SetHealth(nil, 500 * specialCount * ScenarioInfo.Options.opt_General_Difficulty)
					if (specialCount == 1) then
						local satelliteAttack = ForkThread(function(self)
							WaitSeconds(5)
							IssueAggressiveMove({specials[distance][angle].Satellite}, ScenarioUtils.MarkerToPosition('HILL_CENTER'))
							KillThread(self)
						end)
					end

					-- spawn defenses
					spawnDefenses(pos, ScenarioInfo.Options.opt_Defense_Special, "ARMY_SURVIVAL_ENEMY_SPECIALS")

					-- set onkilled function
					specials[distance][angle].OldOnKilled = specials[distance][angle].OnKilled
					specials[distance][angle].myID = distance.." - "..angle
					specials[distance][angle].OnKilled = function(self, instigator, type, overkillRatio)
						--spawn a nuke
						nuke(self:GetPosition())

						-- add points to player

						local instArmy = instigator:GetArmy()
						local armyName = ArmyIndexer(GetArmyBrain(instArmy))
						local scoreMod = math.floor(1 - (GetGameTimeSeconds() / 3600))
						if scoreMod > 0 then
							killCount[armyName] = killCount[armyName] + (100000 * scoreMod)
						end

						-- handle game logic
						msgOut("Special destroyed by "..getUsername(instArmy), 2)
						--LOG(SecondsToClock(GetGameTimeSeconds()).."   Special destroyed by "..getUsername(instArmy))
						self.myID = nil
						self.OldOnKilled(self, instigator, type, overkillRatio)
					end
					foundAngle = true
				else
					angle = math.random(1,4)
				end
			end
		end

		msgOut("Specials spawned for each player", 2)
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Special "..specialCount.." has activated")
		--cinematicsZoom('AREA_CAM_DEFENSE', 2)
		cinematicsZoom('AREA_CAM_OUT', 2)
	end
end

cinematicsZoom = function(area, delay)
	local zoom = ForkThread(function(self)
		Cinematics.EnterNISMode()

		--Cinematics.CameraReset()
		--Cinematics.CameraMoveToRectangle(ScenarioUtils.AreaToRect('AREA_CAM_OUT'), delay)

		--Cinematics.CameraMoveToMarker(ScenarioUtils.GetMarker('HILL_CENTER'), 5)
		--Cinematics.CameraMoveToRectangle(ScenarioUtils.AreaToRect('AREA_CAM_OUT'), 3) -- take 2 sec to zoom out
		--Cinematics.CameraMoveToRectangle(ScenarioUtils.AreaToRect('AREA_CAM_DEFENSE'), 5) -- take 2 sec to zoom in
		Cinematics.CameraMoveToRectangle(ScenarioUtils.AreaToRect(area), delay)
		--WaitSeconds(0.5)

		Cinematics.ExitNISMode()
		KillThread(self)
	end)
end

fireNukes = function()
	if (specialCount > 0) then
		local special1Fired = false
		local special2Fired = false
		local special3Fired = false
		local nukesFired = false
		local uberNukesFired = false
		for i = 1,4 do
			for j = 1,4 do
				if (specials[i][j].myID ~= nil) then
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds()).." : specials[i][j].myID = "..specials[i][j].myID)
					local bp = specials[i][j]:GetBlueprint()
					if (bp ~= nil and table.find(bp.Categories,'NUKE')) then
						if (table.find(bp.Categories,'TECH3')) then
							for k = 1,numNukes do
								specials[i][j]:GiveNukeSiloAmmo(1)
								pos = ScenarioUtils.MarkerToPosition( 'HILL_CENTER' )
								local randAngle = math.random(0,359)
								IssueNuke( { specials[i][j] }, {pos[1] + math.random(1,70) * math.cos(math.rad(randAngle)), pos[2], pos[3] + math.random(1,70) * math.sin(math.rad(randAngle))} )
								nukesFired = true
							end
						else
							for k = 1,numUberNukes do
								specials[i][j]:GiveNukeSiloAmmo(1)
								pos = ScenarioUtils.MarkerToPosition( 'HILL_CENTER' )
								local randAngle = math.random(0,359)
								IssueNuke( { specials[i][j] }, {pos[1] + math.random(1,70) * math.cos(math.rad(randAngle)), pos[2], pos[3] + math.random(1,70) * math.sin(math.rad(randAngle))} )
								uberNukesFired = true
							end
						end
					elseif (bp ~= nil and table.find(bp.Categories,'SIZE16')) then
						local wp = specials[i][j]:GetWeapon(1)
						--wp:ChangeRateOfFire(special3Rate * bp.Weapon[1].RateOfFire)
						wp:ChangeRateOfFire(2)
						special3Fired = true
						--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds()).." : special3 rate = "..bp.Weapon[1].RateOfFire)
					elseif (bp ~= nil and table.find(bp.Categories,'SIZE4')) then
						local wp = specials[i][j]:GetWeapon(1)
						--wp:ChangeRateOfFire(special2Rate * bp.Weapon[1].RateOfFire)
						wp:ChangeRateOfFire(2)
						special2Fired = true
						LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds()).." : special2 rate = "..bp.Weapon[1].RateOfFire)
					elseif (bp ~= nil and table.find(bp.Categories,'ORBITALSYSTEM')) then
						local wp = specials[i][j].Satellite:GetWeapon(1)
						wp:ChangeRateOfFire(special1Rate * specials[i][j].Satellite:GetBlueprint().Weapon[1].RateOfFire)
						special1Fired = true
						--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds()).." : special1 rate = "..specials[i][j].Satellite:GetBlueprint().Weapon[1].RateOfFire)
					end
				end
			end
		end
		if (special1Fired) then
			if (ScenarioInfo.Options.opt_General_WavePeriod > 1) then
				special1Rate = special1Rate + ScenarioInfo.Options.opt_General_WavePeriod
			else
				special1Rate = special1Rate + 1
			end
		end
		if (special2Fired) then
			if (ScenarioInfo.Options.opt_General_WavePeriod > 1) then
				special2Rate = special2Rate + ScenarioInfo.Options.opt_General_WavePeriod
			else
				special2Rate = special2Rate + 1
			end
		end
		if (special3Fired) then
			if (ScenarioInfo.Options.opt_General_WavePeriod > 1) then
				special3Rate = special3Rate + ScenarioInfo.Options.opt_General_WavePeriod
			else
				special3Rate = special3Rate + 1
			end
		end
		if (nukesFired) then
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Incoming nukes per silo: "..numNukes)
			msgOut("Incoming nukes per silo: "..numNukes, 1)
			if (ScenarioInfo.Options.opt_General_WavePeriod > 1) then
				numNukes = numNukes + ScenarioInfo.Options.opt_General_WavePeriod
			else
				numNukes = numNukes + 1
			end
		end
		if (uberNukesFired) then
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Incoming ubernukes per silo: "..numUberNukes)
			msgOut("Incoming ubernukes per silo: "..numUberNukes, 1)
			if (ScenarioInfo.Options.opt_General_WavePeriod > 1) then
				numUberNukes = numUberNukes + ScenarioInfo.Options.opt_General_WavePeriod
			else
				numUberNukes = numUberNukes + 1
			end
		end
	end
end

-- spawns the number of T4s selected in options after a gate dies
spawnGateT4s = function()
	if (ScenarioInfo.Options.opt_Defense_GateDeathT4s < 10) then
		WaitSeconds(1) -- wait for explosion
		for i = 1,8 do
			if (spawnT4[i] == 1) then
				local gateName = "GATE"..i
				local gatePos = ScenarioUtils.MarkerToPosition(gateName)
				local LandT4 = 0
				local AirT4 = 0
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": spawnGateT4s: "..gateName.."   gatePos"..repr(gatePos))

				local maxT4s = 0
				if (ScenarioInfo.Options.opt_Defense_GateDeathT4s == 9) then
					maxT4s = (8 - numGatesOpen)
				else
					maxT4s = ScenarioInfo.Options.opt_Defense_GateDeathT4s
				end
				LandT4 = Utilities.GetRandomInt(1, maxT4s)
				AirT4 = maxT4s - LandT4

				--[[--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": spawnGateT4s: "..gateName)
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": maxT4s: "..maxT4s)
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": LandT4: "..LandT4)
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": AirT4: "..AirT4)
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": NavalT4: "..NavalT4)
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": ")]]--

				local T4SpawnTable = {} -- empty spawn list
				local bpCountL = table.getn(landUnits.Tech4)
				local bpCountA = table.getn(airUnits.Tech4)
				--local bpCountN = table.getn(navalUnits.Tech4)

				for j = 1,LandT4 do -- pull a random entry from the available units table, add it to the spawning table, and repeat for as many units as we will spawn
					local platoonUnits = {} -- create an empty array to combine spawned units to one platoon
					local rr = Utilities.GetRandomInt(1, bpCountL)
					doSpawnUnit(landUnits.Tech4[rr], i, gatePos, platoonUnits, "ARMY_SURVIVAL_ENEMY_SPECIALS")
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Gate Death landUnits.Tech4: "..landUnits.Tech4[rr])
					startAttackMove(platoonUnits, i) --give attack order
				end

				for j = 1,AirT4 do -- pull a random entry from the available units table, add it to the spawning table, and repeat for as many units as we will spawn
					local platoonUnits = {} -- create an empty array to combine spawned units to one platoon
					local rr = Utilities.GetRandomInt(1, bpCountA)
					doSpawnUnit(airUnits.Tech4[rr], i, gatePos, platoonUnits, "ARMY_SURVIVAL_ENEMY_SPECIALS")
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Gate Death airUnits.Tech4: "..airUnits.Tech4[rr])
					startAttackMove(platoonUnits) --give attack order
				end

				spawnT4[i] = nil
			end
		end
	end
end

showArtilleryRange = function()
	local circle = ForkThread(function(self)
		local rangeLand = 165 + (ScenarioInfo.Options.opt_General_Difficulty * 10)
		local pos = ScenarioUtils.MarkerToPosition('ARTILLERY1') -- assuming all Artillery are the same distance from Hill Center
		local artilleryDistance = math.sqrt(math.pow((pos[1] - 512.5), 2) + math.pow((pos[3] - 511.5), 2))
		local artyName
		local segment
		local segmentLand = 2 * math.floor(67.5 - math.deg(math.acos(math.rad(22.5) * artilleryDistance / rangeLand)))
		while (numArtilleryOpen > 0) do
			for i = 1,8 do
				if (artillery[i].myID ~= nil) then
					artyName = 'ARTILLERY'..i
					pos = ScenarioUtils.MarkerToPosition(artyName)
					segment = segmentLand
					range = rangeLand
					for angle = 1,segment do
						DrawLine({pos[1] + (range * math.cos(math.rad(((i - 1) * 45) + (22.5 - segment / 2) + angle - 1))), 26, pos[3] + (range * math.sin(math.rad(((i - 1) * 45) + (22.5 - segment / 2) + angle - 1)))}, {pos[1] + (range * math.cos(math.rad(((i - 1) * 45) + (22.5 - segment / 2) + angle))), 26, pos[3] + (range * math.sin(math.rad(((i - 1) * 45) + (22.5 - segment / 2) + angle)))}, '5F000000')
					end
				end
			end
			WaitSeconds(0.2999)
		end
		KillThread(self)
	end)
end

unitsRemaining = function()
	local numEnemyUnitsPrevious = numEnemyUnits
	local sumEnemyUnitsPrevious = 0
	local sumEnemyUnits = 0

	local catsLand = ((categories.MOBILE - categories.EXPERIMENTAL) * categories.LAND) + categories.uab2302 + categories.urb2302 + categories.ueb2302 + categories.xsb2302 + categories.xab2307 + categories.ueb2401 + categories.url0401 + categories.xsb2305 + categories.ueb2305 + categories.urb2305 + categories.uab2305 + categories.xsb2401
	local catsAir = (categories.MOBILE - categories.EXPERIMENTAL - categories.TRANSPORTATION - categories.CARRIER) * categories.AIR
	local catsNaval = (categories.MOBILE - categories.EXPERIMENTAL - categories.CARRIER - categories.NUKESUB) * categories.NAVAL
	local catsT4 = (categories.MOBILE * categories.EXPERIMENTAL) - categories.url0401
	local catsTransport = categories.TRANSPORTATION
	local catsCarrier = categories.CARRIER
	local catsNukeSub = categories.NUKESUB

	local armySurvivalUnits = {GetArmyBrain("ARMY_SURVIVAL_ENEMY"):GetListOfUnits(catsLand, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY"):GetListOfUnits(catsAir, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY"):GetListOfUnits(catsNaval, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY"):GetListOfUnits(catsT4, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY"):GetListOfUnits(catsTransport, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY"):GetListOfUnits(catsCarrier, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY"):GetListOfUnits(catsNukeSub, false, true)}
	local armySurvivalSpecialUnits = {GetArmyBrain("ARMY_SURVIVAL_ENEMY_SPECIALS"):GetListOfUnits(catsLand, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY_SPECIALS"):GetListOfUnits(catsAir, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY_SPECIALS"):GetListOfUnits(catsNaval, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY_SPECIALS"):GetListOfUnits(catsT4, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY_SPECIALS"):GetListOfUnits(catsTransport, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY_SPECIALS"):GetListOfUnits(catsCarrier, false, true), GetArmyBrain("ARMY_SURVIVAL_ENEMY_SPECIALS"):GetListOfUnits(catsNukeSub, false, true)}

	for i = 1,7 do
		numEnemyUnits[i] = 0
		if (armySurvivalUnits[i] ~= nil) then
			for k,v in ipairs(armySurvivalUnits[i]) do
				if (v:IsDead() == false) then
					numEnemyUnits[i] = numEnemyUnits[i] + 1
				end
			end
		end
		if (armySurvivalSpecialUnits[i] ~= nil) then
			for k,v in ipairs(armySurvivalSpecialUnits[i]) do
				if (v:IsDead() == false) then
					numEnemyUnits[i] = numEnemyUnits[i] + 1
				end
			end
		end
		sumEnemyUnits = sumEnemyUnits + numEnemyUnits[i]
		sumEnemyUnitsPrevious = sumEnemyUnitsPrevious + numEnemyUnitsPrevious[i]
	end

	--[[ don't think this is needed
	if (ScenarioInfo.Options.opt_opt_Defense_Scathis == 0) then
	--if (ScenarioInfo.Options.opt_General_EndGameScenario == 2) then
		numEnemyUnits = numEnemyUnits - numArtilleryOpen
	end
	]]--

	return sumEnemyUnits, sumEnemyUnitsPrevious
	--[[
	if (numGatesOpen == 0) then
		if (sumEnemyUnits >= 100) then
			if (100 * math.ceil(sumEnemyUnits / 100) ~= 100 * math.ceil(sumEnemyUnitsPrevious / 100)) then
				--msgOut("Enemy units: <"..(100 * math.ceil(sumEnemyUnits / 100)), 1)
				msgOut("Enemy units remaining: "..sumEnemyUnits.."\nLand: "..numEnemyUnits[1].."\nAir: "..numEnemyUnits[2].."\nNaval: "..numEnemyUnits[3].."\nT4: "..numEnemyUnits[4], 1)
			end
		elseif (sumEnemyUnits < 100 and sumEnemyUnits >= 25) then
			if (25 * math.ceil(sumEnemyUnits / 25) ~= 25 * math.ceil(sumEnemyUnitsPrevious / 25)) then
				--msgOut("Enemy units: <"..(25 * math.ceil(sumEnemyUnits / 25)), 1)
				msgOut("Enemy units remaining: "..sumEnemyUnits.."\nLand: "..numEnemyUnits[1].."\nAir: "..numEnemyUnits[2].."\nNaval: "..numEnemyUnits[3].."\nT4: "..numEnemyUnits[4], 1)
			end
		elseif (sumEnemyUnits < 25 and sumEnemyUnits >= 5) then
			if (5 * math.ceil(sumEnemyUnits / 5) ~= 5 * math.ceil(sumEnemyUnitsPrevious / 5)) then
				--msgOut("Enemy units: <"..(5 * math.ceil(sumEnemyUnits / 5)), 1)
				msgOut("Enemy units remaining: "..sumEnemyUnits.."\nLand: "..numEnemyUnits[1].."\nAir: "..numEnemyUnits[2].."\nNaval: "..numEnemyUnits[3].."\nT4: "..numEnemyUnits[4], 1)
			end
		else
			if (sumEnemyUnits ~= sumEnemyUnitsPrevious) then
				--msgOut("Enemy units: "..sumEnemyUnits, 1)
				msgOut("Enemy units remaining: "..sumEnemyUnits.."\nLand: "..numEnemyUnits[1].."\nAir: "..numEnemyUnits[2].."\nNaval: "..numEnemyUnits[3].."\nT4: "..numEnemyUnits[4], 1)
			end
		end
	else
		--msgOut("Enemy units: "..sumEnemyUnits, 1)
		msgOut("Enemy units remaining: "..sumEnemyUnits.."\nLand: "..numEnemyUnits[1].."\nAir: "..numEnemyUnits[2].."\nNaval: "..numEnemyUnits[3].."\nT4: "..numEnemyUnits[4], 1)
	end]]--
end

-- activates the superweapons after defenseTime + StartupDelay has elapsed
activateSuperWeapons = function()
	if (waveCount * ScenarioInfo.Options.opt_General_WavePeriod >= (ScenarioInfo.Options.opt_Defense_ObjectStartup) and SWActive == false and ScenarioInfo.Options.opt_Defense_Object < 6) then
		for i = 1,4 do
			if (SWs[i] ~= nil) then
				if (ScenarioInfo.Options.opt_Defense_Object == 1) then
					SWs[i]:SetProductionActive(true)
				elseif (ScenarioInfo.Options.opt_Defense_Object == 5) then

				else
					local wp = SWs[i]:GetWeapon(1)
					wp:SetWeaponEnabled(true)
				end
				SWs[i]:SetCustomName("Activated")
			end
		end
		SWActive = true
		WaitSeconds(0.75)
		spawnSpecials(2)
		if (ScenarioInfo.Options.opt_Defense_Object == 1 and ScenarioInfo.Options.opt_General_Taunts == 1) then
			ScenarioFramework.Dialogue(import('/maps/X1CA_002/X1CA_002_strings.lua').X02_M03_130, nil, true) -- resource gen
		end
	end
end

startupOptions = function()
	local startTemp = {}
	for k,v in pairs(ScenarioInfo.Options) do
		if (string.find(k, 'opt_', 1) ~= nil) then
			local sort = (string.gsub(k, 'opt_', "").." = "..v)
			if (table.getn(startTemp) == 0) then
				table.insert(startTemp, sort)
			else
				local sizeTemp = table.getn(startTemp)
				for index,value in ipairs(startTemp) do
					if (sort < value) then
						table.insert(startTemp, index, sort)
						break
					end
				end
				if (sizeTemp == table.getn(startTemp)) then
					table.insert(startTemp, sort)
				end
			end
		end
	end
	return table.concat(startTemp, "\n")
end

-- timer for spawning waves
spawnThread = function(self)
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Spawn thread running!")

	LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Startup Options:\n\n"..startupOptions().."\n\n")

	if (ScenarioInfo.Options.opt_General_Pause == 1) then
		local responses = {0, 0, 0, 0}
		local humanPlayerCount = 0
		for i,v in ListArmies() do
			if (GetArmyBrain(v).BrainType == 'Human') then
				humanPlayerCount = humanPlayerCount + 1
			end
		end
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": humanPlayerCount = "..humanPlayerCount)
		--msgOut("humanPlayerCount = "..humanPlayerCount)
		local playersNotReady = humanPlayerCount - (responses[1] + responses[2] + responses[3] + responses[4])
		local playersNotReadyNames
		local startupPause1
		--local startupPause2

		local startupPauseThread = ForkThread(function(self)
			for i,v in ipairs(responses) do
				if (v == 0) then
					if (GetArmyBrain(i).BrainType == 'Human') then
						if (playersNotReadyNames ~= nil) then
							playersNotReadyNames = playersNotReadyNames.."\n"..getUsername(i)
						else
							playersNotReadyNames = getUsername(i)
						end
					end
				end
			end
			--startupPause2 = CreateDialogue("Patience please...\n\nPlayers still preparing for battle:\n"..playersNotReadyNames, nil, 'right')
			startupPause1 = CreateDialogue("Each player must click RESUME when ready for the game to begin\n\nPlayers still preparing for battle:\n"..playersNotReadyNames, {'RESUME'}, 'right')

			startupPause1.OnButtonPressed = function(self, button)
				if (button.buttonID == 1) then
					responses[button.presser] = 1
					playersNotReady = humanPlayerCount - (responses[1] + responses[2] + responses[3] + responses[4])
					if (playersNotReady == 0) then
						SimConExecute("ui_lua SessionResume()")
					else
						playersNotReadyNames = nil
						for i,v in ipairs(responses) do
							if (v == 0) then
								if (GetArmyBrain(i).BrainType == 'Human') then
									if (playersNotReadyNames ~= nil) then
										playersNotReadyNames = playersNotReadyNames.."\n"..getUsername(i)
									else
										playersNotReadyNames = getUsername(i)
									end
								end
							end
						end
						startupPause1:SetText("Each player must click RESUME when ready for the game to begin\n\nPlayers still preparing for battle:\n"..playersNotReadyNames)

						--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": dialog = "..repr(startupPause1))
						--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": button = "..repr(button))
						--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": dialog = "..repr(startupPause1.button))
						if (GetFocusArmy() == button.presser) then
							startupPause1:SetButtonDisabled(1, true)
							startupPause1:UpdateButtonText(1, "Patience please...")
							startupPause1:UpdatePosition("center")
						end
					end
				end
			end
			KillThread(self)
		end)

		WaitSeconds(0.89)
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": pausing")
		SimConExecute("ui_lua SessionRequestPause()")
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": after pausing")
		WaitSeconds(1)
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": after pause wait")
		if (startupPause1) then
			startupPause1:Destroy()
		end
		--[[if (startupPause2) then
			startupPause2:Destroy()
		end]]--
	end

	if (ScenarioInfo.Options.opt_General_Taunts == 1) then
		ScenarioFramework.Dialogue(import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M01_030, nil, true) -- glad you're here
	end

	if (ScenarioInfo.Options.opt_General_StartupDelay > 0) then
		msgOut("Waiting "..SecondsToClock(60 * ScenarioInfo.Options.opt_General_StartupDelay).." for startup time to elapse before spawning enemy units", 1)

		-- testing startup delay counter but removed since it got in the way
		--[[local startupDelayThread = ForkThread(function(self)
			local startupDelayDialog = CreateDialogue("Startup Delay Countdown: "..SecondsToClock((ScenarioInfo.Options.opt_General_StartupDelay * 60) - GetGameTimeSeconds()), nil, 'right')
			while (GetGameTimeSeconds() < ScenarioInfo.Options.opt_General_StartupDelay * 60) do
				WaitSeconds(1)
				startupDelayDialog:SetText("Startup Delay Countdown: "..SecondsToClock((ScenarioInfo.Options.opt_General_StartupDelay * 60) - GetGameTimeSeconds()))
			end
			LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": killing dialog")
			startupDelayDialog:Destroy()
			KillThread(self)
		end)]]--
	end

	-- the while loop repeats every spawn interval to spawn a new wave
	while (unitsDead == false) do

		-- wait for next wave
		LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": |__________Waiting "..(ScenarioInfo.Options.opt_General_WavePeriod * 60).."s until next spawn")
		LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": |_________________________________")

		local waitPeriod = (ScenarioInfo.Options.opt_General_StartupDelay * 60) + (waveCount * ScenarioInfo.Options.opt_General_WavePeriod * 60)
		while (GetGameTimeSeconds() < waitPeriod) do
			local waiting = GetGameTimeSeconds()
			if (waitPeriod - waiting > 1) then
				WaitSeconds(math.floor(waitPeriod - waiting))
			else
				WaitSeconds(0.5)
			end
		end
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": done waiting")

		if (ScenarioInfo.Options.opt_General_WavePeriod >= 0.5 or math.mod(waveCount, 60) == 0) then
			taunt()
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": firing nukes")
			fireNukes()
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": nukes fired, SW activation check started")
			activateSuperWeapons()
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": SW activation check completed")
		end
		spawnWave()
	end
	--no more waves to send, kill thread
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Killing spawnThread")
	KillThread(self)
end

taunt = function()
	if (ScenarioInfo.Options.opt_General_Taunts == 1 and numGatesOpen > 0) then
		--ScenarioFramework.Dialogue(OpStrings.X01_M01_240, nil, true) -- enemy subs in north
		--ScenarioFramework.Dialogue(OpStrings.X01_M02_045, nil, true) -- shoot the enemy subs
		local taunts = {
			--'import('/maps/X1CA_TUT/X1CA_TUT_strings.lua').X1T_TU02_080', -- we don't have all day
			--'import('/maps/X1CA_TUT/X1CA_TUT_strings.lua').X1T_TU03_040', -- speed this up
			import('/maps/X1CA_TUT/X1CA_TUT_strings.lua').X1T_TU08_040, -- i want that enemy base destroyed
			--'import('/maps/X1CA_TUT/X1CA_TUT_strings.lua').X1T_TU03_060', -- good job but don't get cocky
			--'import('/maps/X1CA_TUT/X1CA_TUT_strings.lua').X1T_TU06_150', -- maybe we'll make a Commander out of you yet
			import('/maps/X1CA_TUT/X1CA_TUT_strings.lua').X1T_TU08_030, -- wipe out that enemy base

			import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M03_180, -- you looking for a medal?
			import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M01_105, -- quit messing around
			import('/maps/X1CA_001/X1CA_001_strings.lua').X01_M01_106, -- make yourself useful

			import('/maps/X1CA_002/X1CA_002_strings.lua').X02_M02_070, -- multiple attack waves incoming
			import('/maps/X1CA_002/X1CA_002_strings.lua').X02_M02_080, -- another attack
			import('/maps/X1CA_002/X1CA_002_strings.lua').X02_M02_090, -- enemy forces are converging
			import('/maps/X1CA_002/X1CA_002_strings.lua').X02_M02_100, -- enemy assault inbound
			import('/maps/X1CA_002/X1CA_002_strings.lua').X02_M03_230, -- bladder the size of a melon

			import('/maps/X1CA_004/X1CA_004_strings.lua').X04_M03_017, -- Elite Commander, come in
			--'import('/maps/X1CA_004/X1CA_004_strings.lua').X04_M03_040', -- Do whatever it takes

			import('/maps/X1CA_005/X1CA_005_strings.lua').X05_M01_012, -- get it in gear
			import('/maps/X1CA_005/X1CA_005_strings.lua').X05_M03_140, -- attack inbound on your position

			import('/maps/X1CA_006/X1CA_006_strings.lua').X06_DB01_030, -- commander? u there?
			import('/maps/X1CA_006/X1CA_006_strings.lua').X06_M01_140, -- commander?  you read me?
			--'import('/maps/X1CA_006/X1CA_006_strings.lua').X06_M02_013', -- nut has cracked
		}
		if (table.getn(usedTaunts) == table.getn(taunts)) then
			usedTaunts = {}
		end
		local randomTaunt = math.random(1,table.getn(taunts))
		while (table.find(usedTaunts, randomTaunt)) do
			randomTaunt = math.random(1,table.getn(taunts))
		end
		table.insert(usedTaunts, randomTaunt)
		ScenarioFramework.Dialogue(taunts[randomTaunt], nil, true)
	end
end

--[[
showSpecialDestroyArea = function()
	local rectangle = ForkThread(function(self)
		while (true) do
			for i = 1,4 do
				for j = 1,4 do
					pos = {512.5 + (380 + 80 * i) * math.cos(math.rad(135 + 90 * j)), 26, 511.5 + (380 + 80 * i) * math.sin(math.rad(135 + 90 * j))}
					DrawLine({pos[1] - 70, 26, pos[3] - 70}, {pos[1] + 70, 26, pos[3] - 70}, '5F000000')
					DrawLine({pos[1] - 70, 26, pos[3] - 70}, {pos[1] - 70, 26, pos[3] + 70}, '5F000000')
					DrawLine({pos[1] - 70, 26, pos[3] + 70}, {pos[1] + 70, 26, pos[3] + 70}, '5F000000')
					DrawLine({pos[1] + 70, 26, pos[3] + 70}, {pos[1] + 70, 26, pos[3] - 70}, '5F000000')
				end
			end
			WaitSeconds(0.2999)
		end
		KillThread(self)
	end)
end]]--

-- adds the correct number of units for this wave to a table for each gate
spawnWave = function()
	LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": spawnWave() started")

	-- don't do anything if there are no gates left
	if (numGatesOpen > 0) then
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": numGatesOpen: "..numGatesOpen)
		-- find the average gate health
		local sumHealth = 0
		for i = 1,8 do -- sum of gate healths
			if (gates[i].myID ~= nil) then
				sumHealth = sumHealth + gates[i]:GetHealth()
			end
		end
		local aveHealth = (sumHealth/numGatesOpen)

		-- make a table of gates damaged below the average
		local damagedGates = {}
		for i = 1,8 do
			if (gates[i].myID ~= nil) then
				local gateHealthCompare = gates[i]:GetHealth()
				if (gateHealthCompare < aveHealth) then
					table.insert(damagedGates, i)
				end
			end
		end
		local numGatesDamaged = table.getn(damagedGates)
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": numGatesDamaged (below ave): "..numGatesDamaged)

		-- set variables
		local spawnCount = {0, 0, 0, 0, 0, 0, 0}
		local spawnCountLabel = {'Land', 'Air', 'Naval', 'T4', 'Transport', 'Carrier', 'Nuke Sub'}
		local minutesAfterStartupDelay = waveCount * ScenarioInfo.Options.opt_General_WavePeriod
		local waveReverseIndexThisMinute = (1 / ScenarioInfo.Options.opt_General_WavePeriod) - math.mod(waveCount, (1 / ScenarioInfo.Options.opt_General_WavePeriod))
		local minutesAfterPTT = waveCount * ScenarioInfo.Options.opt_General_WavePeriod - ScenarioInfo.Options.opt_General_PeakTechTime

		local landUnitsPerMin = 0
		local landSpawnUnitCount = 0
		local numRemainderLandUnitsThisMinute = 0

		local airUnitsPerMin = 0
		local airSpawnUnitCount = 0
		local numRemainderAirUnitsThisMinute = 0

		local landT4UnitsPerMin = 0
		local airT4UnitsPerMin = 0

		local landT4SpawnUnitCount = 0
		local airT4SpawnUnitCount = 0

		-- Swarm
		if (ScenarioInfo.Options.opt_General_WavePeriod < 1) then
			if (minutesAfterStartupDelay > ScenarioInfo.Options.opt_General_PeakTechTime) then
				landUnitsPerMin = playerCount * ScenarioInfo.Options.opt_NumUnits_Land
				airUnitsPerMin = playerCount * ScenarioInfo.Options.opt_NumUnits_Land
				landT4UnitsPerMin = playerCount * minutesAfterPTT * ScenarioInfo.Options.opt_General_Difficulty / (4 * 10) -- four players on Difficulty 10 would get 1 T4 per min
				airT4UnitsPerMin = playerCount * minutesAfterPTT * ScenarioInfo.Options.opt_General_Difficulty / (3 * 4 * 10) -- four players on Difficulty 10 would get 1/3 T4 per min
			else
				landUnitsPerMin = playerCount * ScenarioInfo.Options.opt_NumUnits_Land * (minutesAfterStartupDelay + 1) / (ScenarioInfo.Options.opt_General_PeakTechTime + 1)
				airUnitsPerMin = playerCount * ScenarioInfo.Options.opt_NumUnits_Air * (minutesAfterStartupDelay + 1) / (ScenarioInfo.Options.opt_General_PeakTechTime + 1)
			end

			landSpawnUnitCount = math.floor(landUnitsPerMin * ScenarioInfo.Options.opt_General_WavePeriod)
			numRemainderLandUnitsThisMinute = landUnitsPerMin - (1 / ScenarioInfo.Options.opt_General_WavePeriod) * landSpawnUnitCount
			if (waveReverseIndexThisMinute <= numRemainderLandUnitsThisMinute) then
				landSpawnUnitCount = landSpawnUnitCount + 1
			end

			landT4SpawnUnitCount = math.floor(landT4UnitsPerMin * ScenarioInfo.Options.opt_General_WavePeriod)
			numRemainderLandT4UnitsThisMinute = landT4UnitsPerMin - (1 / ScenarioInfo.Options.opt_General_WavePeriod) * landT4SpawnUnitCount
			if (waveReverseIndexThisMinute <= numRemainderLandT4UnitsThisMinute) then
				landT4SpawnUnitCount = landT4SpawnUnitCount + 1
			end

			airSpawnUnitCount = math.floor(airUnitsPerMin * ScenarioInfo.Options.opt_General_WavePeriod)
			numRemainderAirUnitsThisMinute = airUnitsPerMin - (1 / ScenarioInfo.Options.opt_General_WavePeriod) * airSpawnUnitCount
			if (waveReverseIndexThisMinute <= numRemainderAirUnitsThisMinute) then
				airSpawnUnitCount = airSpawnUnitCount + 1
			end

			airT4SpawnUnitCount = math.floor(airT4UnitsPerMin * ScenarioInfo.Options.opt_General_WavePeriod)
			numRemainderAirT4UnitsThisMinute = airT4UnitsPerMin - (1 / ScenarioInfo.Options.opt_General_WavePeriod) * airT4SpawnUnitCount
			if (waveReverseIndexThisMinute <= numRemainderAirT4UnitsThisMinute) then
				airT4SpawnUnitCount = airT4SpawnUnitCount + 1
			end

		-- nonSwarm
		else
			if (minutesAfterStartupDelay >= ScenarioInfo.Options.opt_General_PeakTechTime) then
				landUnitsPerMin = playerCount * ScenarioInfo.Options.opt_NumUnits_Land * ScenarioInfo.Options.opt_General_WavePeriod
				airUnitsPerMin = playerCount * ScenarioInfo.Options.opt_NumUnits_Air * ScenarioInfo.Options.opt_General_WavePeriod

				landT4UnitsPerMin = playerCount * (minutesAfterPTT + 1) * ScenarioInfo.Options.opt_General_WavePeriod * ScenarioInfo.Options.opt_General_Difficulty / (4 * 10)
				airT4UnitsPerMin = playerCount * (minutesAfterPTT + 1) * ScenarioInfo.Options.opt_General_WavePeriod * ScenarioInfo.Options.opt_General_Difficulty / (3 * 4 * 10)
			else
				for i = 1,ScenarioInfo.Options.opt_General_WavePeriod do
					if (minutesAfterStartupDelay + 1 - (i - 1) > 0) then
						landUnitsPerMin = landUnitsPerMin + (playerCount * ScenarioInfo.Options.opt_NumUnits_Land * (minutesAfterStartupDelay + 1 - (i - 1)) / (ScenarioInfo.Options.opt_General_PeakTechTime + 1))
						airUnitsPerMin = airUnitsPerMin + (playerCount * ScenarioInfo.Options.opt_NumUnits_Air * (minutesAfterStartupDelay + 1 - (i - 1)) / (ScenarioInfo.Options.opt_General_PeakTechTime + 1))
					end
				end
			end

			landSpawnUnitCount = landUnitsPerMin
			airSpawnUnitCount = airUnitsPerMin

			landT4SpawnUnitCount = landT4UnitsPerMin
			airT4SpawnUnitCount = airT4UnitsPerMin
		end

		local landSpawnUnitDecrement = landSpawnUnitCount
		local airSpawnUnitDecrement = airSpawnUnitCount
		local landT4SpawnUnitDecrement = landT4SpawnUnitCount
		local airT4SpawnUnitDecrement = airT4SpawnUnitCount

		local landSpawn = { {}, {}, {}, {}, {}, {}, {}, {} } -- create an empty array to put each gate's units
		local airSpawn = { {}, {}, {}, {}, {}, {}, {}, {} } -- create an empty array to put each gate's units
		local landT4Spawn = { {}, {}, {}, {}, {}, {}, {}, {} } -- create an empty array to put each gate's units
		local airT4Spawn = { {}, {}, {}, {}, {}, {}, {}, {} } -- create an empty array to put each gate's units

		-- determine how many units exist of each type
		local bpCountLandDFT1 = table.getn(landUnits.Tech1.DirectFire)
		local bpCountLandShieldT1 = table.getn(landUnits.Tech1.Shield)
		local bpCountLandAAT1 = table.getn(landUnits.Tech1.AntiAir)
		local bpCountLandDFT2 = table.getn(landUnits.Tech2.DirectFire)
		local bpCountLandShieldT2 = table.getn(landUnits.Tech2.Shield)
		local bpCountLandAAT2 = table.getn(landUnits.Tech2.AntiAir)
		local bpCountLandDFT3 = table.getn(landUnits.Tech3.DirectFire)
		local bpCountLandShieldT3 = table.getn(landUnits.Tech3.Shield)
		local bpCountLandAAT3 = table.getn(landUnits.Tech3.AntiAir)
		local bpCountLandSpecialT3 = table.getn(landUnits.Tech3.Special)

		local bpCountAirDFT1 = table.getn(airUnits.Tech1.DirectFire)
		local bpCountAirTransportT1 = table.getn(airUnits.Tech1.Transport)
		local bpCountAirAAT1 = table.getn(airUnits.Tech1.AntiAir)
		local bpCountAirDFT2 = table.getn(airUnits.Tech2.DirectFire)
		local bpCountAirTransportT2 = table.getn(airUnits.Tech2.Transport)
		local bpCountAirAAT2 = table.getn(airUnits.Tech2.AntiAir)
		local bpCountAirDFT3 = table.getn(airUnits.Tech3.DirectFire)
		local bpCountAirTransportT3 = table.getn(airUnits.Tech3.Transport)
		local bpCountAirAAT3 = table.getn(airUnits.Tech3.AntiAir)

		local bpCountLandT4 = table.getn(landUnits.Tech4)
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": bpCountLandT4: "..bpCountLandT4)
		local bpCountAirT4 = table.getn(airUnits.Tech4)
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": bpCountAirT4: "..bpCountAirT4)

		-- set ratios
		local landAAT1 = 10
		local landShieldT1 = 20
		if (bpCountLandAAT1 == 0 and bpCountLandShieldT1 == 0) then
			landAAT1 = 0
			landShieldT1 = 0
		elseif (bpCountLandAAT1 == 0) then
			landShieldT1 = landShieldT1 + (landAAT1 / 2)
			landAAT1 = 0
		elseif (bpCountLandShieldT1 == 0) then
			landAAT1 = landAAT1 + (landShieldT1 / 2)
			landShieldT1 = 0
		end

		local landAAT2 = 10
		local landShieldT2 = 20
		if (bpCountLandAAT2 == 0 and bpCountLandShieldT2 == 0) then
			landAAT2 = 0
			landShieldT2 = 0
		elseif (bpCountLandAAT2 == 0) then
			landShieldT2 = landShieldT2 + (landAAT2 / 2)
			landAAT2 = 0
		elseif (bpCountLandShieldT2 == 0) then
			landAAT2 = landAAT2 + (landShieldT2 / 2)
			landShieldT2 = 0
		end

		local landAAT3 = 10
		local landShieldT3 = 20
		if (bpCountLandAAT3 == 0 and bpCountLandShieldT3 == 0) then
			landAAT3 = 0
			landShieldT3 = 0
		elseif (bpCountLandAAT3 == 0) then
			landShieldT3 = landShieldT3 + (landAAT3 / 2)
			landAAT3 = 0
		elseif (bpCountLandShieldT3 == 0) then
			landAAT3 = landAAT3 + (landShieldT3 / 2)
			landShieldT3 = 0
		end

		local airTransportT1 = 5
		local airAAT1 = 10
		if (bpCountAirAAT1 == 0) then
			airAAT1 = 0
		end
		if (bpCountAirTransportT1 == 0) then
			airTransportT1 = 0
		end

		local airTransportT2 = 5
		local airAAT2 = 10
		if (bpCountAirAAT2 == 0) then
			airAAT2 = 0
		end
		if (bpCountAirTransportT2 == 0) then
			airTransportT2 = 0
		end

		local airTransportT3 = 5
		local airAAT3 = 10
		if (bpCountAirAAT3 == 0) then
			airAAT3 = 0
		end
		if (bpCountAirTransportT3 == 0) then
			airTransportT3 = 0
		end

		-- choose random gate to start the spawn cycle from
		local randomLandGate = math.random(1,8)
		local randomAirGate = math.random(1,8)

		--------------------
		-- LAND SPAWN

		-- cycle through gates and add one land unit to that gate's table until all number of units to spawn has been reached
		for i = 1,landSpawnUnitCount do
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": land spawn loop")
			for j = 1,8 do
				local s = 1 + math.mod(j + randomLandGate, 8)
				if (gates[s].myID ~= nil) then
					local tech1ratio = 100 - (250 * minutesAfterStartupDelay/ScenarioInfo.Options.opt_General_PeakTechTime)
					local tech2ratio = 200 - (250 * minutesAfterStartupDelay/ScenarioInfo.Options.opt_General_PeakTechTime)
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": tech2ratio: "..tech2ratio)
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": minutesAfterStartupDelay: "..minutesAfterStartupDelay)
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": ScenarioInfo.Options.opt_General_PeakTechTime: "..ScenarioInfo.Options.opt_General_PeakTechTime)
					local tech3ratio = 300 - (250 * minutesAfterStartupDelay/ScenarioInfo.Options.opt_General_PeakTechTime)
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": tech3ratio: "..tech3ratio)
					local supcomratio = 100 - tech3ratio
					local t4ratio = 0

					if (tech1ratio > 0) then
						tech2ratio = 100 - tech1ratio
						tech3ratio = 0
						supcomratio = 0
					else
						tech1ratio = 0
						if (tech2ratio > 0) then
							tech3ratio = 100 - tech2ratio
							supcomratio = 0
						else
							tech2ratio = 0
							if (tech3ratio < 50) then
								tech3ratio = 50
								t4ratio = 100
							end
						end
					end
					local addUnit = 1
					if (table.find(damagedGates, s)) then -- triple the units going to gates damaged below the average
						addUnit = 3
					end
					for l = 1, addUnit do
						if (landSpawnUnitDecrement - 1 >= 0) then
							local rand1 = Utilities.GetRandomInt(1, 100)
							local rand2 = Utilities.GetRandomInt(1, 100)
							if (rand1 <= tech1ratio) then
								--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": LAND rand1 <= tech1ratio")
								if (rand2 <= landAAT1) then
									local rr = Utilities.GetRandomInt(1, bpCountLandAAT1)
									table.insert(landSpawn[s], landUnits.Tech1.AntiAir[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landUnits.Tech1.AntiAir: "..landUnits.Tech1.AntiAir[rr])
								elseif (rand2 <= landAAT1 + landShieldT1) then
									local rr = Utilities.GetRandomInt(1, bpCountLandShieldT1)
									table.insert(landSpawn[s], landUnits.Tech1.Shield[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landUnits.Tech1.Shield: "..landUnits.Tech1.Shield[rr])
								else
									local rr = Utilities.GetRandomInt(1, bpCountLandDFT1)
									table.insert(landSpawn[s], landUnits.Tech1.DirectFire[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landUnits.Tech1.DirectFire: "..landUnits.Tech1.DirectFire[rr])
								end
							elseif (rand1 > 100 - tech2ratio) then
								--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": LAND rand1 > 100 - tech2ratio")
								if (rand2 <= landAAT2) then
									local rr = Utilities.GetRandomInt(1, bpCountLandAAT2)
									table.insert(landSpawn[s], landUnits.Tech2.AntiAir[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landUnits.Tech2.AntiAir: "..landUnits.Tech2.AntiAir[rr])
								elseif (rand2 <= landAAT2 + landShieldT2) then
									local rr = Utilities.GetRandomInt(1, bpCountLandShieldT2)
									table.insert(landSpawn[s], landUnits.Tech2.Shield[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landUnits.Tech2.Shield: "..landUnits.Tech2.Shield[rr])
								else
									local rr = Utilities.GetRandomInt(1, bpCountLandDFT2)
									table.insert(landSpawn[s], landUnits.Tech2.DirectFire[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landUnits.Tech2.DirectFire: "..landUnits.Tech2.DirectFire[rr])
								end
							elseif (rand1 <= tech3ratio) then
								--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": LAND rand1 < tech3ratio")
								if (rand2 <= landAAT3) then
									local rr = Utilities.GetRandomInt(1, bpCountLandAAT3)
									table.insert(landSpawn[s], landUnits.Tech3.AntiAir[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landUnits.Tech3.AntiAir: "..landUnits.Tech3.AntiAir[rr])
								elseif (rand2 <= landAAT3 + landShieldT3) then
									local rr = Utilities.GetRandomInt(1, bpCountLandShieldT3)
									table.insert(landSpawn[s], landUnits.Tech3.Shield[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landUnits.Tech3.Shield: "..landUnits.Tech3.Shield[rr])
								else
									local rr = Utilities.GetRandomInt(1, bpCountLandDFT3)
									table.insert(landSpawn[s], landUnits.Tech3.DirectFire[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landUnits.Tech3.DirectFire: "..landUnits.Tech3.DirectFire[rr])
								end
							else
								--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": LAND rand1 else supcom")
								local rr = Utilities.GetRandomInt(1, bpCountLandSpecialT3)
								table.insert(landSpawn[s], landUnits.Tech3.Special[rr])
								--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landUnits.Tech3.Special: "..landUnits.Tech3.Special[rr])
							end
							landSpawnUnitDecrement = landSpawnUnitDecrement - 1
						end
						if (minutesAfterStartupDelay >= ScenarioInfo.Options.opt_General_PeakTechTime) then
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": building land T4 spawn table")
							--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": landT4SpawnUnitDecrement: "..landT4SpawnUnitDecrement)
							if (landT4SpawnUnitDecrement - 1 >= 0) then
								local rr = Utilities.GetRandomInt(1, bpCountLandT4)
								table.insert(landT4Spawn[s], landUnits.Tech4[rr])
								landT4SpawnUnitDecrement = landT4SpawnUnitDecrement - 1
							end
						end
					end
				end
			end
		end

		-- spawn land units at each gate
		for i = 1,8 do
			if (gates[i].myID ~= nil) then
				local platoonUnits = {} -- create an empty array to combine land units at each gate into one platoon
				local gateName = "GATE"..i
				pos = ScenarioUtils.MarkerToPosition(gateName)
				for index, id in landSpawn[i] do -- iterate through the list of ids we have just generated and spawn the units
					doSpawnUnit(id, i, pos, platoonUnits, "ARMY_SURVIVAL_ENEMY")
					spawnCount[1] = spawnCount[1] + 1
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": doSpawn loop, gate = "..i)
				end
				startAttackMove(platoonUnits, i) --give attack order (send as platoon - not individually)
				platoonUnits = {} -- create an empty array to combine land units at each gate into one platoon
				for index, id in landT4Spawn[i] do -- iterate through the list of ids we have just generated and spawn the units
					doSpawnUnit(id, i, pos, platoonUnits, "ARMY_SURVIVAL_ENEMY")
					spawnCount[4] = spawnCount[4] + 1
					startAttackMove(platoonUnits) --give attack order (send individually - not in platoon)
				end
			end
		end

		--------------------
		-- AIR SPAWN

		-- cycle through gates and add one air unit to that gate's table until all number of units to spawn has been reached
		--if (math.floor(waveCount / 3) == (waveCount / 3)) then
		for i = 1,airSpawnUnitCount do
			for j = 1,8 do
				local s = 1 + math.mod(j + randomAirGate, 8)
				if (gates[s].myID ~= nil) then

					tech1ratio = 100 - (200 * minutesAfterStartupDelay/ScenarioInfo.Options.opt_General_PeakTechTime)
					tech2ratio = 200 - (200 * minutesAfterStartupDelay/ScenarioInfo.Options.opt_General_PeakTechTime)
					tech3ratio = 100 - tech2ratio

					if (tech1ratio > 0) then
						tech2ratio = 100 - tech1ratio
						tech3ratio = 0
					else
						tech1ratio = 0
						if (tech2ratio < 0) then
							tech2ratio = 0
							tech3ratio = 100
						end
					end
					local addUnit = 1
					if (table.find(damagedGates, s)) then -- triple the units going to gates damaged below the average
						addUnit = 3
					end
					for l = 1, addUnit do
						if (airSpawnUnitDecrement - 1 >= 0) then
							local rand1 = Utilities.GetRandomInt(1, 100)
							local rand2 = Utilities.GetRandomInt(1, 100)
							if (rand1 <= tech1ratio) then
								--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": AIR rand1 <= tech1ratio")
								if (rand2 <= airTransportT1) then
									local rr = Utilities.GetRandomInt(1, bpCountAirTransportT1)
									table.insert(airSpawn[s], airUnits.Tech1.Transport[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": airUnits.Tech1.Transport: "..airUnits.Tech1.Transport[rr])
									spawnCount[5] = spawnCount[5] + 1
								elseif (rand2 <= airTransportT1 + airAAT1) then
									local rr = Utilities.GetRandomInt(1, bpCountAirAAT1)
									table.insert(airSpawn[s], airUnits.Tech1.AntiAir[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": airUnits.Tech1.AntiAir: "..airUnits.Tech1.AntiAir[rr])
								else
									local rr = Utilities.GetRandomInt(1, bpCountAirDFT1)
									table.insert(airSpawn[s], airUnits.Tech1.DirectFire[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": airUnits.Tech1.DirectFire: "..airUnits.Tech1.DirectFire[rr])
								end
							elseif (rand1 > 100 - tech2ratio) then
								--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": AIR rand1 > 100 - tech2ratio")
								if (rand2 <= airTransportT2) then
									local rr = Utilities.GetRandomInt(1, bpCountAirTransportT2)
									table.insert(airSpawn[s], airUnits.Tech2.Transport[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": airUnits.Tech2.Transport: "..airUnits.Tech2.Transport[rr])
									spawnCount[5] = spawnCount[5] + 1
								elseif (rand2 <= airTransportT2 + airAAT2) then
									local rr = Utilities.GetRandomInt(1, bpCountAirAAT2)
									table.insert(airSpawn[s], airUnits.Tech2.AntiAir[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": airUnits.Tech2.AntiAir: "..airUnits.Tech2.AntiAir[rr])
								else
									local rr = Utilities.GetRandomInt(1, bpCountAirDFT2)
									table.insert(airSpawn[s], airUnits.Tech2.DirectFire[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": airUnits.Tech2.DirectFire: "..airUnits.Tech2.DirectFire[rr])
								end
							else
								--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": AIR rand1 else tech3")
								if (rand2 <= airTransportT3) then
									local rr = Utilities.GetRandomInt(1, bpCountAirTransportT3)
									table.insert(airSpawn[s], airUnits.Tech3.Transport[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": airUnits.Tech3.Transport: "..airUnits.Tech3.Transport[rr])
									spawnCount[5] = spawnCount[5] + 1
								elseif (rand2 <= airTransportT3 + airAAT3) then
									local rr = Utilities.GetRandomInt(1, bpCountAirAAT3)
									table.insert(airSpawn[s], airUnits.Tech3.AntiAir[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": airUnits.Tech3.AntiAir: "..airUnits.Tech3.AntiAir[rr])
								else
									local rr = Utilities.GetRandomInt(1, bpCountAirDFT3)
									table.insert(airSpawn[s], airUnits.Tech3.DirectFire[rr])
									--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": airUnits.Tech3.DirectFire: "..airUnits.Tech3.DirectFire[rr])
								end
							end
							airSpawnUnitDecrement = airSpawnUnitDecrement - 1
						end
						if (minutesAfterStartupDelay >= ScenarioInfo.Options.opt_General_PeakTechTime) then
							if (airT4SpawnUnitDecrement - 1 >= 0) then
								local rr = Utilities.GetRandomInt(1, bpCountAirT4)
								table.insert(airT4Spawn[s], airUnits.Tech4[rr])
								airT4SpawnUnitDecrement = airT4SpawnUnitDecrement - 1
							end
						end
					end
				end
			end
		end

		-- spawn units at each gate
		for i = 1,8 do
			if (gates[i].myID ~= nil) then
				local platoonUnits = {} -- create an empty array to combine land units at each gate into one platoon
				local gateName = "GATE"..i
				pos = ScenarioUtils.MarkerToPosition(gateName)
				for index, id in airSpawn[i] do -- iterate through the list of ids we have just generated and spawn the units
					doSpawnUnit(id, i, pos, platoonUnits, "ARMY_SURVIVAL_ENEMY")
					spawnCount[2] = spawnCount[2] + 1
					startAttackMove(platoonUnits) --give attack order (send individually - not in platoon)
				end
				platoonUnits = {} -- create an empty array to combine land units at each gate into one platoon
				for index, id in airT4Spawn[i] do -- iterate through the list of ids we have just generated and spawn the units
					doSpawnUnit(id, i, pos, platoonUnits, "ARMY_SURVIVAL_ENEMY")
					spawnCount[4] = spawnCount[4] + 1
					startAttackMove(platoonUnits) --give attack order (send individually - not in platoon)
				end
			end
		end
		spawnCount[2] = spawnCount[2] - spawnCount[5] - spawnCount[6]

		sumEnemyUnits, sumEnemyUnitsPrevious = unitsRemaining()
		local spawnCountTotal = 0
		for i = 1,7 do
			spawnCountTotal = spawnCountTotal + spawnCount[i]
		end

		local waveMessage
		if (spawnCountTotal > 0) then
			if (ScenarioInfo.Options.opt_General_WavePeriod >= 0.5) then
				waveMessage = "Attack wave "..(waveCount + 1).." has spawned\nTotal: "..spawnCountTotal.." / "..sumEnemyUnits
				local focusArmy = GetFocusArmy()
				if (focusArmy == -1) then
					focusArmy = 5
				end
				if (dialogSettings[focusArmy][1] == 3) then
					for i = 1,7 do
						if (numEnemyUnits[i] > 0) then
							waveMessage = waveMessage.."\n"..spawnCountLabel[i]..": "..spawnCount[i].." / "..numEnemyUnits[i]
						end
					end
					msgOut(waveMessage, 2)
				else
					msgOut(waveMessage, 1)
				end
				LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": \n"..waveMessage.."\n\n")
			end
		end

		if (math.abs(minutesAfterStartupDelay - ScenarioInfo.Options.opt_General_PeakTechTime) < 1/60) then
			msgOut("Peak Tech Time reached", 2)
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": T4 spawning initiated")
		end
	end
	waveCount = waveCount + 1
	LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": spawnWave() ended")
end

-- does the actual spawning of the units
doSpawnUnit = function(id, gate, pos, platoon, army)
	spawnAngle = spawnAngle + 170
	if (spawnAngle > 360) then
		spawnAngle = spawnAngle - 360
	end

	local unit
	for i = 1,3 do
		if (airUnits['Tech'..i].Transport and table.find(airUnits['Tech'..i].Transport, id)) then
			unit = CreateUnitHPR(id, "ARMY_SURVIVAL_ENEMY_SPECIALS", pos[1] + 45 * math.cos(math.rad(spawnAngle)), pos[2], pos[3] + 45 * math.sin(math.rad(spawnAngle)),0,0,0)
		end
	end
	if (not unit) then
		--unit = CreateUnitHPR(id, "ARMY_SURVIVAL_ENEMY", pos[1] + 45 * math.cos(math.rad(spawnAngle)), pos[2], pos[3] + 45 * math.sin(math.rad(spawnAngle)),0,0,0)
		unit = CreateUnitHPR(id, army, pos[1] + 45 * math.cos(math.rad(spawnAngle)), pos[2], pos[3] + 45 * math.sin(math.rad(spawnAngle)),0,0,0)
	end

	if (unit == nil) then
		return
	end

	-- no wreckage to speed up game... can be removed when we all have quadcores/octacores
	local bp = unit:GetBlueprint()
	if (bp ~= nil) then
		bp.Wreckage = nil
	end

	local veterancyLand = math.floor(ScenarioInfo.Options.opt_Veterancy_Land * waveCount * ScenarioInfo.Options.opt_General_WavePeriod / ScenarioInfo.Options.opt_General_PeakTechTime)
	local veterancyAir = math.floor(ScenarioInfo.Options.opt_Veterancy_Air * waveCount * ScenarioInfo.Options.opt_General_WavePeriod / ScenarioInfo.Options.opt_General_PeakTechTime)
	if (not unit.AddLevels) then
		if (veterancyLand > 5) then
			veterancyLand = 5
		end
		if (veterancyAir > 5) then
			veterancyAir = 5
		end
		--[[if (veterancyNaval > 5) then
			veterancyNaval = 5
		end]]--
	end

	--give SupComs a random enhancement and apply healthBuff to everyone else based on elapsed time
	if (id == "ual0301") then		-- aeon
		local ucnt = table.getn(upgradeTable["A"])
		local enh = upgradeTable["A"][Utilities.GetRandomInt(1, ucnt)]
		unit:CreateEnhancement(enh)
	elseif (id == "url0301") then	-- cybran
		local ucnt = table.getn(upgradeTable["R"])
		local enh = upgradeTable["R"][Utilities.GetRandomInt(1, ucnt)]
		if (enh == 'StealthGenerator') then	-- stealth needs cloaking first, bug in game
			unit:CreateEnhancement('CloakingGenerator')
		end
		unit:CreateEnhancement(enh)
	elseif (id == "uel0301") then	-- uef
		local ucnt = table.getn(upgradeTable["E"])
		local enh = upgradeTable["E"][Utilities.GetRandomInt(1, ucnt)]
		unit:CreateEnhancement(enh)
	elseif (id == "xsl0301") then	-- seraphim
		local ucnt = table.getn(upgradeTable["S"])
		local enh = upgradeTable["S"][Utilities.GetRandomInt(1, ucnt)]
		unit:CreateEnhancement(enh)
	else
		--apply buffs but not to SupComs or T4s
		if (not table.find(bp.Categories, "EXPERIMENTAL")) then
			if table.find(bp.Categories, "AIR") then
				unit:SetMaxHealth(unit:GetHealth() * (1 + (ScenarioInfo.Options.opt_HealthBuffer_Air - 1) * waveCount * ScenarioInfo.Options.opt_General_WavePeriod / ScenarioInfo.Options.opt_General_PeakTechTime))
				unit:SetHealth(nil, unit:GetHealth() * (1 + (ScenarioInfo.Options.opt_HealthBuffer_Air - 1) * waveCount * ScenarioInfo.Options.opt_General_WavePeriod / ScenarioInfo.Options.opt_General_PeakTechTime))
				if (unit.AddLevels) then
					unit:AddLevels(veterancyAir)
				else
					unit:SetVeterancy(veterancyAir)
				end
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": MaxHealth: "..(1 + (ScenarioInfo.Options.opt_healthBuffer - 1) * (waveCount - 1) * ScenarioInfo.Options.opt_General_WavePeriod / ScenarioInfo.Options.opt_General_PeakTechTime))
			elseif table.find(bp.Categories, "LAND") then
				unit:SetMaxHealth(unit:GetHealth() * (1 + (ScenarioInfo.Options.opt_HealthBuffer_Land - 1) * waveCount * ScenarioInfo.Options.opt_General_WavePeriod / ScenarioInfo.Options.opt_General_PeakTechTime))
				unit:SetHealth(nil, unit:GetHealth() * (1 + (ScenarioInfo.Options.opt_HealthBuffer_Land - 1) * waveCount * ScenarioInfo.Options.opt_General_WavePeriod / ScenarioInfo.Options.opt_General_PeakTechTime))
				if (unit.AddLevels) then
					--LOG("_____LOG: AddLevels found")
					unit:AddLevels(veterancyLand)
				else
					--LOG("_____LOG: AddLevels not found")
					unit:SetVeterancy(veterancyLand)
				end
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": MaxHealth: "..(1 + (ScenarioInfo.Options.opt_healthBuffer - 1) * (waveCount - 1) * ScenarioInfo.Options.opt_General_WavePeriod / ScenarioInfo.Options.opt_General_PeakTechTime))
			end
		end
	end

	-- pull out transports and carriers and fill them up, and pull out nukesubs too
	if EntityCategoryContains(categories.TECH1 * categories.TRANSPORTATION, unit) then
		local level = 1
		local transportThread = ForkThread(function(self)
			doTransport(unit, level, veterancyLand)
			KillThread(self)
		end)
	elseif EntityCategoryContains(categories.TECH2 * categories.TRANSPORTATION, unit) then
		local level = 2
		local transportThread = ForkThread(function(self)
			doTransport(unit, level, veterancyLand)
			KillThread(self)
		end)
	elseif EntityCategoryContains(categories.TECH3 * categories.TRANSPORTATION, unit) then
		local level = 3
		local transportThread = ForkThread(function(self)
			doTransport(unit, level, veterancyLand)
			KillThread(self)
		end)
	--[[elseif EntityCategoryContains(categories.TRANSPORTATION * categories.EXPERIMENTAL, unit) then
		local level = 1
		local transportThread = ForkThread(function(self)
			doTransport(unit, level, veterancyLand)
			KillThread(self)
		end)]]--
	else
		-- insert to platoon list
		table.insert(platoon, unit)
	end
end

randomOpenGate = function(type)
	local randomGate = 0
	while (randomGate == 0) do
		local rand = 0
		if (type == 'land') then
			rand = Utilities.GetRandomInt(1,8)
			if (gates[rand].myID ~= nil) then
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": found a gate2")
				randomGate = rand
			elseif (gates[1].myID == nil and gates[2].myID == nil and gates[3].myID == nil and gates[4].myID == nil and gates[5].myID == nil and gates[6].myID == nil and gates[7].myID == nil and gates[8].myID == nil) then
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": land gates dead, breaking loop2")
				break
			end
		else
			rand = Utilities.GetRandomInt(1,8)
			if (gates[rand].myID ~= nil) then
				--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": found a gate3")
				randomGate = rand
			elseif (numGatesOpen == 0) then
				break
			end
		end
	end
	return randomGate
end

-- loads transports with the correct number of units and gives the orders to the units
doTransport = function(transport, level, veterancyLand)
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": starting doTransport")
	while (not transport:IsDead()) do -- while it is still alive, keep loading transport at a random gate and unloading at a random waypoint
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": started while loop")
		-- load transport
		local position = transport:GetPosition()
		local unit = {'ual0103', 'xsl0202', 'xel0305', 'url0301'}
		local unitsToTransport = {}

		local minutesAfterStartupDelay = waveCount * ScenarioInfo.Options.opt_General_WavePeriod
		local landTransportUnitCount = ScenarioInfo.Options.opt_NumUnits_Land * (minutesAfterStartupDelay + 1) / (ScenarioInfo.Options.opt_General_PeakTechTime + 1)
		if (landTransportUnitCount > ScenarioInfo.Options.opt_NumUnits_Land) then
			landTransportUnitCount = ScenarioInfo.Options.opt_NumUnits_Land
		end
		if (landTransportUnitCount == 0) then
			startAttackMove({transport})
			return
		end
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": spawning units to transport")
		for i = 1,landTransportUnitCount do
			local rand = Utilities.GetRandomInt(1, table.getn(landUnits['Tech'..level].DirectFire))
			unitsToTransport[i] = CreateUnitHPR(landUnits['Tech'..level].DirectFire[rand], "ARMY_SURVIVAL_ENEMY_SPECIALS", position.x, 2 * position.y, position.z, 0,0,0)
			unitsToTransport[i]:SetMaxHealth(unitsToTransport[i]:GetHealth() * (1 + (ScenarioInfo.Options.opt_HealthBuffer_Land - 1) * waveCount * ScenarioInfo.Options.opt_General_WavePeriod / ScenarioInfo.Options.opt_General_PeakTechTime))
			unitsToTransport[i]:SetHealth(nil, unitsToTransport[i]:GetHealth() * (1 + (ScenarioInfo.Options.opt_HealthBuffer_Land - 1) * waveCount * ScenarioInfo.Options.opt_General_WavePeriod / ScenarioInfo.Options.opt_General_PeakTechTime))
			if (unitsToTransport[i].AddLevels) then
				unitsToTransport[i]:AddLevels(veterancyLand)
			else
				unitsToTransport[i]:SetVeterancy(veterancyLand)
			end
		end
		ScenarioFramework.AttachUnitsToTransports(unitsToTransport, {transport})
		--local unitsDestroyed = 0
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": removing extra units")
		for i = 1,landTransportUnitCount do
			if (not unitsToTransport[i]:IsUnitState('Attached')) then
				IssueDestroySelf({unitsToTransport[i]})
				--unitsDestroyed = unitsDestroyed + 1
			end
		end
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": extra units destroyed")
		-- give unload order
		IssueTransportUnload({transport}, ScenarioUtils.MarkerToPosition("WP"..Utilities.GetRandomInt(1,4)))
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": transport ordered to unload")
		IssueAggressiveMove(unitsToTransport, ScenarioUtils.MarkerToPosition('HILL_CENTER'))
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": units ordered to attack center")
		--[[
		-- wait until the transport has arrived at it's destination and has completed unloading
		local attachedUnits = 55555
		while (attachedUnits > 0) do
			WaitSeconds(5)
			attachedUnits = table.getn(transport:GetAttachedUnitsList())
		end]]--

		-- find a random open gate and send the transport for reload
		local randomGate = randomOpenGate()
		IssueMove({transport}, ScenarioUtils.MarkerToPosition("GATE"..randomGate))
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Transport ordered to move to random gate")

		-- wait until the transport has reached the gate
		local distanceToGate = 55555
		while (distanceToGate > 45) do
			WaitSeconds(1)
			-- if gate has died then choose another
			if (gates[randomGate].myID == nil) then
				IssueClearCommands({ transport })
				randomGate = randomOpenGate('land')
				if (randomGate == 0) then
					startAttackMove({transport})
					return
				else
					IssueMove({transport}, ScenarioUtils.MarkerToPosition("GATE"..randomGate))
				end
			end
			-- find distance to gate
			local transportPosition = transport:GetPosition()
			local gatePosition = ScenarioUtils.MarkerToPosition("GATE"..randomGate)
			distanceToGate = math.sqrt(math.pow((transportPosition[1] - gatePosition[1]), 2) + math.pow((transportPosition[3] - gatePosition[3]), 2))
		end
		IssueClearCommands({ transport })
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Transport has reached random gate")
	end
end

-- gives units attack orders
startAttackMove = function(units, gateNum)	--forms a platoon to attack the hill with a given set of units
	if (table.getn(units) > 0) then
		local aiBrain = GetArmyBrain("ARMY_SURVIVAL_ENEMY")
		local attackPlatoon = aiBrain:MakePlatoon('','')
		local fcnt = table.getn(formationTable)
		local formation = formationTable[Utilities.GetRandomInt(1, fcnt)]
		aiBrain:AssignUnitsToPlatoon(attackPlatoon, units, 'Attack', formation)
		attackPlatoon:AggressiveMoveToLocation(ScenarioUtils.MarkerToPosition("WP"..Utilities.GetRandomInt(1,8)))
		attackPlatoon:Patrol(ScenarioUtils.MarkerToPosition('HILL_CENTER'))
		attackPlatoon:Patrol(ScenarioUtils.MarkerToPosition("WP"..Utilities.GetRandomInt(1,8)))
	end
end

------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- misc functions
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

-- spawn acu and kill it to create a fake nuke explosion :=) *hack hack hack, but the gpg devs said it's easiest way atm*
nuke = function(position)
	local acu = CreateUnitHPR( "ual0001", "ARMY_SURVIVAL_ENEMY", position[1], position[2], position[3], 0,0,0)
	if (acu ~= nil) then
		acu:Kill()
	end
end

-- kills all commanders of a specified army
killAllCommanders = function(army)
	local list = GetArmyBrain(army):GetListOfUnits(categories.COMMAND , false)
	for i, u in list do
		if (u:IsDead() == false) then
			u:Kill()
		end
	end
end

--given a player returns a proper username
getUsername = function(army)
	return GetArmyBrain(army).Nickname
end

--given an ai brain returns an army index
function ArmyIndexer(aiBrain)
	local army = aiBrain:GetArmyIndex()
	for i, v in ipairs(ListArmies()) do
	--for i, v in ListArmies() do
		if (i == army) then
			return v
		end
	end
end

-- calculates scores
armyKills = function(brain)
	local kills = brain:GetArmyStat('Enemies_Killed',0.0).Value
	return kills
end

dialogThread = function(self)
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Dialog thread running!")
	local focusArmy = GetFocusArmy()
	if (focusArmy == -1) then
		focusArmy = 5
	end
	local detailButton = {'Disable Dialogs', 'Detail:Low', 'Detail: High'}
	local buttonButton = {'Buttons: Show', 'Buttons: Hide'}
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": focusArmy = "..focusArmy)
	while (unitsDead == false) do
		if (table.getn(allPlayerDialogs[focusArmy]) ~= 0) then
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": message found")
			local buttons
			if (dialogSettings[focusArmy][3]) then
				buttons = {detailButton[(dialogSettings[focusArmy][1])], 'Time: '..dialogSettings[focusArmy][2]..'s', 'Buttons: Show'}
			end
			local dialog = table.concat(allPlayerDialogs[focusArmy], "\n\n")
			allPlayerDialogs[focusArmy] = {}
			local myDialog = CreateDialogue(dialog, buttons, 'right')
			myDialog.OnButtonPressed = function(self, button)
				if (button.buttonID == 1) then
					dialogSettings[button.presser][1] = 1 + math.mod(dialogSettings[button.presser][1], 3)
					if (focusArmy == button.presser) then
						myDialog:UpdateButtonText(1, detailButton[(dialogSettings[focusArmy][1])])
					end
				elseif (button.buttonID == 2) then
					dialogSettings[button.presser][2] = 5 + math.mod(dialogSettings[button.presser][2], 60)
					if (focusArmy == button.presser) then
						myDialog:UpdateButtonText(2, 'Time: '..dialogSettings[focusArmy][2].."s")
					end
				elseif (button.buttonID == 3) then
					dialogSettings[button.presser][3] = not dialogSettings[button.presser][3]
					if (focusArmy == button.presser) then
						if (dialogSettings[focusArmy][3]) then
							myDialog:UpdateButtonText(3, 'Buttons: Show')
						else
							myDialog:UpdateButtonText(3, 'Buttons: Hide')
						end
					end
				end
			end
			--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": message displayed")
			local numSecondsWaited = 0
			local secondsToWait = dialogSettings[focusArmy][2]
			while (numSecondsWaited <= secondsToWait and unitsDead == false) do
				WaitSeconds(1)
				if table.getn(allPlayerDialogs[focusArmy]) ~= 0 then
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": another message found")
					buttons = nil
					if (dialogSettings[focusArmy][3]) then
						buttons = {detailButton[(dialogSettings[focusArmy][1])], 'Time: '..dialogSettings[focusArmy][2]..'s', 'Buttons: Show'}
					end
					dialog = table.concat(allPlayerDialogs[focusArmy], "\n\n")
					allPlayerDialogs[focusArmy] = {}
					myDialog:Destroy()
					WaitSeconds(0.25)
					local myDialog = CreateDialogue(dialog, buttons, 'right')
					myDialog.OnButtonPressed = function(self, button)
						if (button.buttonID == 1) then
							dialogSettings[button.presser][1] = math.mod(dialogSettings[button.presser][1] + 1, 3)
							if (focusArmy == button.presser) then
								myDialog:UpdateButtonText(1, detailButton[(dialogSettings[focusArmy][1])])
							end
						elseif (button.buttonID == 2) then
							dialogSettings[button.presser][2] = 5 + math.mod(dialogSettings[button.presser][2], 60)
							if (focusArmy == button.presser) then
								myDialog:UpdateButtonText(2, 'Time: '..dialogSettings[focusArmy][2].."s")
							end
						elseif (button.buttonID == 3) then
							dialogSettings[button.presser][3] = not dialogSettings[button.presser][3]
							if (focusArmy == button.presser) then
								if (dialogSettings[focusArmy][3]) then
									myDialog:UpdateButtonText(3, 'Buttons: Show')
								else
									myDialog:UpdateButtonText(3, 'Buttons: Hide')
								end
							end
						end
					end
					numSecondsWaited = 0
					--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": message updated")
				else
					numSecondsWaited = numSecondsWaited + 1
				end
			end
			myDialog:Destroy()
		end
		WaitSeconds(1)
	end
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": Killing dialogThread")
	KillThread(self)
end

-- display a message to all players
msgOut = function(text, verbosity)
	local focusArmy = GetFocusArmy()
	if (focusArmy == -1) then
		focusArmy = 5
	end
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": focus army > 0")
	--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": dialogSettings.focusArmy[1] = "..dialogSettings[focusArmy][1])
	if (dialogSettings[focusArmy][1] == 3) then
		table.insert(allPlayerDialogs[focusArmy], 1, text)
		--LOG("_____LOG: "..SecondsToClock(GetGameTimeSeconds())..": text inserted")
	elseif (dialogSettings[focusArmy][1] == 2) then
		if (verbosity == 1) then
			table.insert(allPlayerDialogs[focusArmy], 1, text)
		end
	end
	for i,v in ipairs(allPlayerDialogs[focusArmy]) do
		if (i > 5) then
			table.remove(allPlayerDialogs[focusArmy], i)
		end
	end
end

-- coverts seconds to hh:mm:ss format
SecondsToClock = function(sSeconds)
	local nSeconds = tonumber(sSeconds)
	if (nSeconds == 0) then
		--return nil
		return "00:00:00"
	else
		local nHours = math.floor(nSeconds/3600)
		local nMins = math.floor((nSeconds/60) - (nHours * 60))
		local nSecs = math.floor(nSeconds - nHours * 3600 - nMins * 60)
		nHours = string.format("%02.f", nHours)
		nMins = string.format("%02.f", nMins)
		nSecs = string.format("%02.f", nSecs)
		return nHours..":"..nMins..":"..nSecs
	end
end

--EOF
